# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/DataTime.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//DataTime
"""

from os import linesep
from sys import getsizeof
from ._check import set_array, check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from .DataND import DataND

# Import all class method
# Try/catch to remove unnecessary dependencies in unused method
try:
    from ..Methods.DataTime.time_to_freq import time_to_freq
except ImportError as error:
    time_to_freq = error


from numpy import array, array_equal
from ._check import InitUnKnowClassError


class DataTime(DataND):
    """Class for fields defined in time space"""

    VERSION = 1

    # cf Methods.DataTime.time_to_freq
    if isinstance(time_to_freq, ImportError):
        time_to_freq = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use DataTime method time_to_freq: " + str(time_to_freq)
                )
            )
        )
    else:
        time_to_freq = time_to_freq
    # save and copy methods are available in all object
    save = save
    copy = copy

    def __init__(
        self,
        axes=None,
        FTparameters=-1,
        values=None,
        is_real=True,
        symbol="",
        name="",
        unit="",
        normalizations=-1,
        init_dict=None,
        init_str=None,
    ):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionnary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "axes" in list(init_dict.keys()):
                axes = init_dict["axes"]
            if "FTparameters" in list(init_dict.keys()):
                FTparameters = init_dict["FTparameters"]
            if "values" in list(init_dict.keys()):
                values = init_dict["values"]
            if "is_real" in list(init_dict.keys()):
                is_real = init_dict["is_real"]
            if "symbol" in list(init_dict.keys()):
                symbol = init_dict["symbol"]
            if "name" in list(init_dict.keys()):
                name = init_dict["name"]
            if "unit" in list(init_dict.keys()):
                unit = init_dict["unit"]
            if "normalizations" in list(init_dict.keys()):
                normalizations = init_dict["normalizations"]
        # Set the properties (value check and convertion are done in setter)
        # Call DataND init
        super(DataTime, self).__init__(
            axes=axes,
            FTparameters=FTparameters,
            values=values,
            is_real=is_real,
            symbol=symbol,
            name=name,
            unit=unit,
            normalizations=normalizations,
        )
        # The class is frozen (in DataND init), for now it's impossible to
        # add new properties

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        DataTime_str = ""
        # Get the properties inherited from DataND
        DataTime_str += super(DataTime, self).__str__()
        return DataTime_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False

        # Check the properties inherited from DataND
        if not super(DataTime, self).__eq__(other):
            return False
        return True

    def compare(self, other, name="self"):
        """Compare two objects and return list of differences"""

        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()

        # Check the properties inherited from DataND
        diff_list.extend(super(DataTime, self).compare(other, name=name))
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object

        # Get size of the properties inherited from DataND
        S += super(DataTime, self).__sizeof__()
        return S

    def as_dict(self, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        # Get the properties inherited from DataND
        DataTime_dict = super(DataTime, self).as_dict(**kwargs)
        # The class name is added to the dict for deserialisation purpose
        # Overwrite the mother class name
        DataTime_dict["__class__"] = "DataTime"
        return DataTime_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        # Set to None the properties inherited from DataND
        super(DataTime, self)._set_None()



# -*- coding: utf-8 -*-
from SciDataTool.Functions.Plot import unit_dict

import csv
import numpy as np
from os.path import join


def export_along(
    self,
    *args,
    unit="SI",
    is_norm=False,
    axis_data=[],
    save_path=None,
    file_name=None,
    file_format="csv",
    is_multiple_files=False,
):
    """Exports the sliced or interpolated version of the data, using conversions and symmetries if needed, in a file.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    file_format: str
        export format ("csv", ...)
    Returns
    -------
    a DataND object
    """

    # If args not specified, write only oneperiod/pattern
    if args == tuple():
        # Dynamic import to avoid loop
        module = __import__("SciDataTool.Classes.DataPattern", fromlist=["DataPattern"])
        DataPattern = getattr(module, "DataPattern")
        arg_list = []
        for axis in self.get_axes():
            if isinstance(axis, DataPattern):
                arg_list.append(axis.name + "[pattern]")
            else:
                arg_list.append(axis.name + "[oneperiod]")
        args = tuple(arg_list)

    # Get requested data
    is_fft = False
    for arg in args:
        if "freqs" in arg or "wavenumber" in arg:
            is_fft = True
    if is_fft:
        results = self.get_magnitude_along(
            *args, unit=unit, is_norm=is_norm, axis_data=axis_data
        )
    else:
        results = self.get_along(*args, unit=unit, is_norm=is_norm, axis_data=axis_data)
    axes_list = results["axes_list"]

    # Remove slice axes
    axes_list_new = []
    slices = ""
    for axis in axes_list:
        if axis.unit == "SI":
            axis.unit = unit_dict[axis.name]
        if len(results[axis.name]) == 1:
            slices += axis.name + "=" + str(results[axis.name][0])
        else:
            axes_list_new.append(axis)
    for axis in results["axes_dict_other"]:
        if slices != "":
            slices = slices + ", "
        slices += axis + "=" + str(results["axes_dict_other"][axis][0])
    if slices != "":
        slices = "sliced at " + slices

    # Default file_name
    if file_name is None:
        file_name = self.symbol + "_Data"

    if file_format == "csv":
        # Write csv files
        # Format: first axis in column, second in row, third in file
        if len(axes_list_new) == 3 and is_multiple_files:
            nfiles = len(axes_list_new[2].values)
        elif len(axes_list_new) == 3:
            raise Exception(
                "ERROR: cannot export more than 2 dimensions in single csv file. Activate is_multiple_files to write in multiple csv files."
            )
        elif len(axes_list_new) < 3:
            nfiles = 1
        else:
            raise Exception("ERROR: cannot export more than 3 dimensions in csv file")

        for i in range(nfiles):
            if nfiles > 1:
                file_name_i = (
                    file_name
                    + "_"
                    + axes_list_new[2].name
                    + str(axes_list_new[2].values[i])
                )
                if slices == "":
                    slices = "sliced at "
                slices_i = (
                    slices
                    + axes_list_new[2].name
                    + "="
                    + str(axes_list_new[2].values[i])
                )
            else:
                file_name_i = file_name
                slices_i = slices
            with open(
                join(save_path, file_name_i + "." + file_format), "w+", newline=""
            ) as my_csv:
                csvWriter = csv.writer(my_csv, delimiter=",")
                # First line: meta-data
                if unit == "SI":
                    unit = self.unit
                meta_data = [self.symbol, self.name, "[" + unit + "]", slices_i]
                csvWriter.writerow(meta_data)

                # Second line: axes + second axis values
                if len(axes_list_new) == 1:
                    A2_cell = axes_list_new[0].name + "[" + axes_list_new[0].unit + "]"
                    second_line = [A2_cell]
                else:
                    A2_cell = (
                        axes_list_new[0].name
                        + "["
                        + axes_list_new[0].unit
                        + "]"
                        + "/"
                        + axes_list_new[1].name
                        + "["
                        + axes_list_new[1].unit
                        + "]"
                    )
                    second_line = np.insert(
                        results[axes_list_new[1].name].astype("object"), 0, A2_cell
                    )
                csvWriter.writerow(second_line)

                # Rest of file: first axis + matrix
                if len(results[self.symbol].shape) == 1:
                    # Transpose if 1D array
                    field = results[self.symbol].T
                elif nfiles > 1:
                    # Slice third axis
                    field = np.take(results[self.symbol], i, axis=2)
                else:
                    field = results[self.symbol]
                matrix = np.column_stack((results[axes_list_new[0].name].T, field))
                csvWriter.writerows(matrix)

    else:
        raise Exception("ERROR: export format not supported")



# -*- coding: utf-8 -*-
from SciDataTool.Functions.parser import read_input_strings
from SciDataTool.Functions.fft_functions import comp_fftn, comp_ifftn


def get_along(
    self,
    *args,
    unit="SI",
    is_norm=False,
    axis_data=[],
    is_squeeze=True,
    is_magnitude=False
):
    """Returns the ndarray of the field, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    Returns
    -------
    list of 1Darray of axes values, ndarray of field values
    """
    # Read the axes input in args
    if len(args) == 1 and type(args[0]) == tuple:
        args = args[0]  # if called from another script with *args
    axes_list = read_input_strings(args, axis_data)
    # Extract the requested axes (symmetries + unit)
    axes_list, transforms = self.comp_axes(axes_list)
    # Get the field
    values = self.get_field(axes_list)
    # If DataFreq + 1 ifft axis + 1 fft axis: perform ifft on all axes then fft
    save_transforms = None
    save_names = None
    if "ifft" in transforms and "fft" in transforms:
        save_transforms = [axis.transform for axis in axes_list]
        save_names = [axis.name for axis in axes_list]
        for axis in axes_list:
            if axis.name == "freqs":
                axis.transform = "ifft"
                axis.name = "time"
            elif axis.name == "wavenumber":
                axis.transform = "ifft"
                axis.name = "angle"
    # Inverse fft
    if "ifft" in transforms:
        values = comp_ifftn(values, axes_list, is_real=self.is_real)
    # Prepare fft in ifft/fft case
    if save_transforms is not None:
        for i, transform in enumerate(save_transforms):
            axes_list[i].name = save_names[i]
            if transform == "fft_axis":
                axes_list[i].transform = "fft"
                save_transforms[i] = "fft"
            else:
                axes_list[i].transform = transform
    # Slices along time/space axes
    values, axes_dict_other = self.extract_slices(values, axes_list)
    # fft
    if "fft" in transforms:
        values = comp_fftn(values, axes_list, is_real=self.is_real)
    # Slices along fft axes
    values = self.extract_slices_fft(values, axes_list)
    # Rebuild symmetries
    values = self.rebuild_symmetries(values, axes_list)
    # Interpolate over axis values
    values = self.interpolate(values, axes_list)
    # Conversions
    values = self.convert(values, unit, is_norm, is_squeeze, is_magnitude, axes_list)
    # Return axes and values
    return_dict = {}
    for axis_requested in axes_list:
        if axis_requested.extension in ["sum", "rss", "mean", "rms", "integrate"]:
            return_dict[axis_requested.name] = axis_requested.extension
        else:
            return_dict[axis_requested.name] = axis_requested.values
    return_dict[self.symbol] = values
    return_dict["axes_list"] = axes_list
    return_dict["axes_dict_other"] = axes_dict_other
    return return_dict



from logging import getLogger
from typing import Dict, List

import numpy as np
from numpy import all as np_all
from numpy import allclose, exp, linspace
from numpy import max as np_max
from numpy import mean
from numpy import min as np_min
from numpy import moveaxis, ndarray, outer, pi, tensordot, where


def matrice_D(t: ndarray, f: ndarray) -> ndarray:
    """
    Construct discrete Fourier transform matrix
    """
    return exp(-1j * outer(f * 2 * pi, t))


def nudftn(a: ndarray, axes_dict: Dict[int, List[ndarray]]) -> ndarray:
    """
    Compute the non uniform discrete Fourier Transform

    Parameter
    ---------

    a: Input data
    axes_dict: Axes to perform dft with corresponding space and freqsampling

    Returns
    res: DFT result
    """
    # Copy data
    res = a.copy()

    axes_fft = []
    # Iterate on each axe to compute 1D DFT
    for idx_axe, axes in axes_dict.items():
        if idx_axe not in axes_fft:  # Check if DFT has to be performed
            # Extract space sampling and frequency sampling
            s, f = axes

            # Check aliasing with criteria frequency < 1/2*mean(spacestep)
            # Other criteria may be used with min instead of mean : max(f)>(1/(2*min(space)))
            space = s[1:] - s[:-1]
            if np_max(f) > (1 / (2 * mean(space))):
                # Display warning
                print(
                    f"The maximum frequency ({np_max(f)}) is greater than the 1/(2*mean of each timestep) ({1/(2*mean(space))}), aliasing may occur."
                )

            # TODO criteria to have only positive freq or approximatively 50% of negative frequencies

            # Construct DFT matrix
            D = matrice_D(s, f)

            # Operate the matrix multiplication and move axis to keep the good shaê
            res = tensordot(D, res, ((1,), (idx_axe)))
            res = moveaxis(res, 0, idx_axe)

            # Normalize by the number of frequencies
            res /= len(s)

            # Check if every frequency is >=0 (rfft equivalent)
            if np_all(f >= 0):
                idx = where(f != 0)
                slice_list = tuple(
                    slice(None) if i != idx_axe else idx[0] for i in range(res.ndim)
                )
                res[slice_list] *= 2

    return res


def matrice_E(t: ndarray, f: ndarray) -> ndarray:
    """Construct inverse discrete Fourier transform matrix"""

    return exp(1j * outer(t, f * 2 * pi))


def inudftn(a: ndarray, axes_dict: Dict[int, List[ndarray]]) -> ndarray:
    """
    Compute the non uniform discrete Fourier Transform

    Parameter
    ---------

    a: Input data
    axes_dict: Axes to perform dft with corresponding space and freqsampling

    Returns
    res: DFT result
    """

    res = a.copy()

    # TODO check axes to perform ifft if space and freq sampling are matching
    # /!\ Normalization is done in inudft and not in fft, look in SciDataTool to find the
    # right normalization
    axes_fft = []

    # Iterate on each axe to compute 1D DFT
    for idx_axe, axes in axes_dict.items():
        if idx_axe not in axes_fft:  # Check if DFT has to be performed
            # Extract space sampling and frequency sampling
            s, f = axes

            # Construct DFT matrix
            E = matrice_E(s, f)

            # Operate the matrix multiplication and move axis to keep the good shape
            res = tensordot(E, res, ((1,), (idx_axe)))
            res = moveaxis(res, 0, idx_axe)

    return res


def is_uniform(vect_a, rtol=1e-05, atol=1e-8):
    """
    Tests if the array vect_a is uniform
    Parameters
    ----------
    vect_a: array
        Time or space vector
    rtol: float
        Relative tolerance parameter
    atol: float
        Absolut tolerance parameter
    Returns
    -------
    y: boolean
    boolean True if absolute(a - b) <= (atol + rtol * absolute(b))
    """

    assert type(vect_a) == type(ndarray(shape=())), TypeError(
        "Time or space vector given must be a ndarray"
    )
    n = len(vect_a)

    minimum = np_min(vect_a)
    maximum = np_max(vect_a)

    y = allclose(
        vect_a,
        linspace(minimum, maximum, n),
        rtol=rtol,
        atol=atol,
        equal_nan=False,
    )

    return y



# -*- coding: utf-8 -*-


def get_periodicity(self):
    """Gives periodicity of the axis.
    Parameters
    ----------
    self: Data1D
        a Data1D object
    Returns
    -------
    per, is_antiper
    """

    per = 1
    is_antiper = False

    if "antiperiod" in self.symmetries:
        if self.symmetries["antiperiod"] > 1:
            per = self.symmetries["antiperiod"]
            is_antiper = True
    elif "period" in self.symmetries:
        if self.symmetries["period"] > 1:
            per = self.symmetries["period"]

    return (per, is_antiper)



# -*- coding: utf-8 -*-
from SciDataTool.Functions import AxisError


def get_axis_periodic(self, Nper, is_antiperiod=False):
    """Returns the vector 'axis' taking symmetries into account.
    Parameters
    ----------
    self: Data1D
        a Data1D object
    Nper: int
        number of periods
    is_antiperiod: bool
        return values on a semi period (only for antiperiodic signals)
    Returns
    -------
    New Data1D
    """

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.Data1D", fromlist=["Data1D"])
    Data1D = getattr(module, "Data1D")

    values = self.values
    N = self.get_length()

    if N % Nper != 0:
        raise AxisError(
            "ERROR: length of axis is not divisible by the number of periods"
        )

    values_per = values[: int(N / Nper)]

    if is_antiperiod:
        sym = "antiperiod"
    else:
        sym = "period"

    New_axis = Data1D(
        values=values_per,
        name=self.name,
        unit=self.unit,
        symmetries={sym: Nper},
        normalizations=self.normalizations,
        is_components=self.is_components,
        symbol=self.symbol,
    )

    return New_axis



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/VectorField.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//VectorField
"""

from os import linesep
from sys import getsizeof
from ._check import check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from ._frozen import FrozenClass

# Import all class method
# Try/catch to remove unnecessary dependencies in unused method
try:
    from ..Methods.VectorField.freq_to_time import freq_to_time
except ImportError as error:
    freq_to_time = error

try:
    from ..Methods.VectorField.get_axes import get_axes
except ImportError as error:
    get_axes = error

try:
    from ..Methods.VectorField.get_harm_rphiz_along import get_harm_rphiz_along
except ImportError as error:
    get_harm_rphiz_along = error

try:
    from ..Methods.VectorField.get_harm_xyz_along import get_harm_xyz_along
except ImportError as error:
    get_harm_xyz_along = error

try:
    from ..Methods.VectorField.get_mag_rphiz_along import get_mag_rphiz_along
except ImportError as error:
    get_mag_rphiz_along = error

try:
    from ..Methods.VectorField.get_mag_xyz_along import get_mag_xyz_along
except ImportError as error:
    get_mag_xyz_along = error

try:
    from ..Methods.VectorField.get_rphiz_along import get_rphiz_along
except ImportError as error:
    get_rphiz_along = error

try:
    from ..Methods.VectorField.get_xyz_along import get_xyz_along
except ImportError as error:
    get_xyz_along = error

try:
    from ..Methods.VectorField.plot_2D_Data import plot_2D_Data
except ImportError as error:
    plot_2D_Data = error

try:
    from ..Methods.VectorField.plot_2D_Data_Animated import plot_2D_Data_Animated
except ImportError as error:
    plot_2D_Data_Animated = error

try:
    from ..Methods.VectorField.plot_3D_Data import plot_3D_Data
except ImportError as error:
    plot_3D_Data = error

try:
    from ..Methods.VectorField.time_to_freq import time_to_freq
except ImportError as error:
    time_to_freq = error


from ._check import InitUnKnowClassError


class VectorField(FrozenClass):
    """Class for 2D or 3D vector fields (time or frequency domain)"""

    VERSION = 1

    # Check ImportError to remove unnecessary dependencies in unused method
    # cf Methods.VectorField.freq_to_time
    if isinstance(freq_to_time, ImportError):
        freq_to_time = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method freq_to_time: " + str(freq_to_time)
                )
            )
        )
    else:
        freq_to_time = freq_to_time
    # cf Methods.VectorField.get_axes
    if isinstance(get_axes, ImportError):
        get_axes = property(
            fget=lambda x: raise_(
                ImportError("Can't use VectorField method get_axes: " + str(get_axes))
            )
        )
    else:
        get_axes = get_axes
    # cf Methods.VectorField.get_harm_rphiz_along
    if isinstance(get_harm_rphiz_along, ImportError):
        get_harm_rphiz_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method get_harm_rphiz_along: "
                    + str(get_harm_rphiz_along)
                )
            )
        )
    else:
        get_harm_rphiz_along = get_harm_rphiz_along
    # cf Methods.VectorField.get_harm_xyz_along
    if isinstance(get_harm_xyz_along, ImportError):
        get_harm_xyz_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method get_harm_xyz_along: "
                    + str(get_harm_xyz_along)
                )
            )
        )
    else:
        get_harm_xyz_along = get_harm_xyz_along
    # cf Methods.VectorField.get_mag_rphiz_along
    if isinstance(get_mag_rphiz_along, ImportError):
        get_mag_rphiz_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method get_mag_rphiz_along: "
                    + str(get_mag_rphiz_along)
                )
            )
        )
    else:
        get_mag_rphiz_along = get_mag_rphiz_along
    # cf Methods.VectorField.get_mag_xyz_along
    if isinstance(get_mag_xyz_along, ImportError):
        get_mag_xyz_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method get_mag_xyz_along: "
                    + str(get_mag_xyz_along)
                )
            )
        )
    else:
        get_mag_xyz_along = get_mag_xyz_along
    # cf Methods.VectorField.get_rphiz_along
    if isinstance(get_rphiz_along, ImportError):
        get_rphiz_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method get_rphiz_along: "
                    + str(get_rphiz_along)
                )
            )
        )
    else:
        get_rphiz_along = get_rphiz_along
    # cf Methods.VectorField.get_xyz_along
    if isinstance(get_xyz_along, ImportError):
        get_xyz_along = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method get_xyz_along: " + str(get_xyz_along)
                )
            )
        )
    else:
        get_xyz_along = get_xyz_along
    # cf Methods.VectorField.plot_2D_Data
    if isinstance(plot_2D_Data, ImportError):
        plot_2D_Data = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method plot_2D_Data: " + str(plot_2D_Data)
                )
            )
        )
    else:
        plot_2D_Data = plot_2D_Data
    # cf Methods.VectorField.plot_2D_Data_Animated
    if isinstance(plot_2D_Data_Animated, ImportError):
        plot_2D_Data_Animated = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method plot_2D_Data_Animated: "
                    + str(plot_2D_Data_Animated)
                )
            )
        )
    else:
        plot_2D_Data_Animated = plot_2D_Data_Animated
    # cf Methods.VectorField.plot_3D_Data
    if isinstance(plot_3D_Data, ImportError):
        plot_3D_Data = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method plot_3D_Data: " + str(plot_3D_Data)
                )
            )
        )
    else:
        plot_3D_Data = plot_3D_Data
    # cf Methods.VectorField.time_to_freq
    if isinstance(time_to_freq, ImportError):
        time_to_freq = property(
            fget=lambda x: raise_(
                ImportError(
                    "Can't use VectorField method time_to_freq: " + str(time_to_freq)
                )
            )
        )
    else:
        time_to_freq = time_to_freq
    # save and copy methods are available in all object
    save = save
    copy = copy
   

    def __init__(self, name="", symbol="", components=-1, init_dict = None, init_str = None):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionnary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "name" in list(init_dict.keys()):
                name = init_dict["name"]
            if "symbol" in list(init_dict.keys()):
                symbol = init_dict["symbol"]
            if "components" in list(init_dict.keys()):
                components = init_dict["components"]
        # Set the properties (value check and convertion are done in setter)
        self.parent = None
        self.name = name
        self.symbol = symbol
        self.components = components

        # The class is frozen, for now it's impossible to add new properties
        self._freeze()

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        VectorField_str = ""
        if self.parent is None:
            VectorField_str += "parent = None " + linesep
        else:
            VectorField_str += "parent = " + str(type(self.parent)) + " object" + linesep
        VectorField_str += 'name = "' + str(self.name) + '"' + linesep
        VectorField_str += 'symbol = "' + str(self.symbol) + '"' + linesep
        VectorField_str += "components = "+ str(self.components) + linesep + linesep
        return VectorField_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False
        if other.name != self.name:
            return False
        if other.symbol != self.symbol:
            return False
        if other.components != self.components:
            return False
        return True

    def compare(self, other, name='self'):
        """Compare two objects and return list of differences"""

        if type(other) != type(self):
            return ['type('+name+')']
        diff_list = list()
        if other._name != self._name:
            diff_list.append(name+'.name')
        if other._symbol != self._symbol:
            diff_list.append(name+'.symbol')
        if (other.components is None and self.components is not None) or (other.components is not None and self.components is None):
            diff_list.append(name+'.components None mismatch')
        elif self.components is None:
            pass
        elif len(other.components) != len(self.components):
            diff_list.append('len('+name+'components)')
        else:
            for key in self.components:
                diff_list.extend(self.components[key].compare(other.components[key],name=name+'.components'))
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object
        S += getsizeof(self.name)
        S += getsizeof(self.symbol)
        if self.components is not None:
            for key, value in self.components.items():
                S += getsizeof(value) + getsizeof(key)
        return S

    def as_dict(self, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        Optional keyword input parameter is for internal use only 
        and may prevent json serializability.
        """

        VectorField_dict = dict()
        VectorField_dict["name"] = self.name
        VectorField_dict["symbol"] = self.symbol
        if self.components is None:
            VectorField_dict["components"] = None
        else:
            VectorField_dict["components"] = dict()
            for key, obj in self.components.items():
                if obj is not None:
                    VectorField_dict["components"][key] = obj.as_dict()
                else:
                    VectorField_dict["components"][key] = None
        # The class name is added to the dict for deserialisation purpose
        VectorField_dict["__class__"] = "VectorField"
        return VectorField_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        self.name = None
        self.symbol = None
        self.components = None

    def _get_name(self):
        """getter of name"""
        return self._name

    def _set_name(self, value):
        """setter of name"""
        check_var("name", value, "str")
        self._name = value

    name = property(
        fget=_get_name,
        fset=_set_name,
        doc=u"""Name of the vector field

        :Type: str
        """,
    )

    def _get_symbol(self):
        """getter of symbol"""
        return self._symbol

    def _set_symbol(self, value):
        """setter of symbol"""
        check_var("symbol", value, "str")
        self._symbol = value

    symbol = property(
        fget=_get_symbol,
        fset=_set_symbol,
        doc=u"""Symbol of the vector field

        :Type: str
        """,
    )

    def _get_components(self):
        """getter of components"""
        if self._components is not None:
            for key, obj in self._components.items():
                if obj is not None:
                    obj.parent = self
        return self._components

    def _set_components(self, value):
        """setter of components"""
        if type(value) is dict:
            for key, obj in value.items():
                if type(obj) is dict:
                    class_obj = import_class('SciDataTool.Classes', obj.get('__class__'), 'components')
                    value[key] = class_obj(init_dict=obj)
        if type(value) is int and value == -1:
            value = dict()
        check_var("components", value, "{DataND}")
        self._components = value

    components = property(
        fget=_get_components,
        fset=_set_components,
        doc=u"""Dict of the components

        :Type: {SciDataTool.Classes.DataND}
        """,
    )



# -*- coding: utf-8 -*-


def has_period(self):
    """Indicates if an axis has symmetries.
    Parameters
    ----------
    self: DataPattern
        a DataPatternobject
    Returns
    -------
    Boolean
    """

    return False



# -*- coding: utf-8 -*-

from numpy import array, empty, int32, squeeze, expand_dims


def set_array(obj, prop, value):
    """Set an array that can be None or a list
    Parameters
    ----------
    obj : ?
        Object to set
    prop : str
        Name of the property to set
    value : ?
        Value to set
    Returns
    -------
    None
    """
    if value is None:  # Default value
        value = empty(0)
    elif isinstance(value, list):
        value = array(value)
    setattr(obj, prop, value)


def check_init_dict(init_dict, key_list):
    """Check if init_dict is correct to initialize the Object (all the keys of init_dict are in key_list)
    Parameters
    ----------
    init_dict : dict
        The dictionary to check
    key_list : list
        List of the expected key (str)
    Returns
    -------
    None
    Raises
    ------
    NotADictError
        init_dict is not a dictionary
    MissingInitDictKeyError
        A key is missing in init_dict
    """
    if not isinstance(init_dict, dict):
        raise NotADictError("Init by dict : init_dict must be a dict")
    msg = "Second argument is the list of variable needed in init_dict"
    assert isinstance(key_list, list), msg
    # Check that every key of init_dict is expected for this object
    key_list.append("__class__")  # Not a property, added for load
    for key in list(init_dict.keys()):
        if key not in key_list:
            raise UnknowInitDictKeyError("Init by dict : " + key + " is not a property")


def check_var(var_name, value, expect_type, Vmin=None, Vmax=None):
    """Check if var_name can be set with value
    Parameters
    ----------
    var_name : str
        The name of the property to set
    value : ?
        The value to check
    expected_type : str
        The name of the expected type for value
    Vmin : float
        Value must be >Vmin (None if the property has no Vmin) (Default value = None)
    Vmax : float
        Value must be <Vmax (None if the property has no Vmax) (Default value = None)
    Returns
    -------
    None
    Raises
    ------
    CheckError
        The value is incorrect for var_name
    """
    if value is not None:
        type_value = type(value).__name__
        if type_value == "float64":
            type_value = "float"
        # Check if value has the good type
        check_type(var_name, value, expect_type, type_value)
        # Check if value is in the good range (if needed)
        if Vmin is not None:
            check_min(var_name, value, type_value, Vmin)
        if Vmax is not None:
            check_max(var_name, value, type_value, Vmax)


def check_type(var_name, value, expect_type, type_value):
    """Check if value has the expected type for var_name
    Parameters
    ----------
    var_name : str
        The name of the property to set
    value : ?
        The value to check
    expected_type : str
        The name of the expected type for value (str)
    type_value :
        The name of the actual type of the variable
    expect_type :
    Returns
    -------
    Raises
    ------
    CheckTypeError
        Value has a wrong type for var_nam
    """
    if expect_type == "float":  # float variable can take int value
        if not (
            (isinstance(value, int32) and not isinstance(value, bool))
            or (isinstance(value, int) and not isinstance(value, bool))
            or isinstance(value, float)
        ):
            raise CheckTypeError(
                "For " + var_name + " : Double expected, " + type_value + " given"
            )
    elif expect_type == "complex":
        # complex variable can take float or int value
        if not (
            (isinstance(value, int) and not isinstance(value, bool))
            or isinstance(value, float)
            or isinstance(value, complex)
        ):
            raise CheckTypeError(
                "For " + var_name + " : Complex expected, " + type_value + " given"
            )
    elif expect_type == "str":
        if not isinstance(value, str):
            raise CheckTypeError(
                "For " + var_name + " : String expected, " + type_value + " given"
            )
    elif expect_type == "list":
        if not isinstance(value, list):
            raise CheckTypeError(
                "For " + var_name + " : list expected, " + type_value + " given"
            )
    elif expect_type == "int" and type_value == "float":
        if value % 1 != 0:  # If not an Integer
            raise CheckTypeError(
                "For " + var_name + " : Integer expected, " + type_value + " given"
            )
    elif expect_type == "int" and type_value == "int32":
        if value % 1 != 0:  # If not an Integer
            raise CheckTypeError(
                "For " + var_name + " : Integer expected, " + type_value + " given"
            )
    elif expect_type[0] == "[" and expect_type[-1] == "]":  # List of type
        if not isinstance(value, list):
            raise CheckTypeError(
                "For " + var_name + " : List expected, " + type_value + " given"
            )
        # Check type of value element
        expect_type = expect_type[1:-1]
        for element in value:
            type_value = type(element).__name__
            # Check if it's the expected type
            if not type_value == expect_type:
                # Check if object inherit from the expected type
                mother = element.__class__.__bases__  # List of inherited class
                find = False  # Store the result of the research
                while mother != ():  # Every class inherit from FrozenClass
                    if mother[0].__name__ == expect_type:
                        find = True
                        break
                    else:
                        mother = mother[0].__bases__
                if not find:  # Not the good type
                    raise CheckTypeError(
                        "For "
                        + var_name
                        + " : "
                        + expect_type
                        + " expected, "
                        + type_value
                        + " given"
                    )
    elif expect_type[0] == "{" and expect_type[-1] == "}":  # Dict of type
        if not isinstance(value, dict):
            raise CheckTypeError(
                "For " + var_name + " : Dict expected, " + type_value + " given"
            )
        # Check type of value element
        expect_type = expect_type[1:-1]
        for key, element in value.items():
            type_value = type(element).__name__
            # Check if it's the expected type
            if not type_value == expect_type:
                # Check if object inherit from the expected type
                mother = element.__class__.__bases__  # List of inherited class
                find = False  # Store the result of the research
                while mother != ():  # Every class inherit from FrozenClass
                    if mother[0].__name__ == expect_type:
                        find = True
                        break
                    else:
                        mother = mother[0].__bases__
                if not find:  # Not the good type
                    raise CheckTypeError(
                        "For "
                        + var_name
                        + " : "
                        + expect_type
                        + " expected, "
                        + type_value
                        + " given"
                    )
    else:
        if not type_value == expect_type:  # Check if it's the expected type
            if expect_type in ["int", "bool", "str"]:
                # We don't check inheritance for python type
                raise CheckTypeError(
                    "For "
                    + var_name
                    + " : "
                    + expect_type
                    + " expected, "
                    + type_value
                    + " given"
                )
            else:
                # Check if object inherit from the expected type
                mother = value.__class__.__bases__  # List of inherited class
                find = False  # Store the result of the research
                while mother != ():  # Every class inherit from FrozenClass
                    if mother[0].__name__ == expect_type:
                        find = True
                        break
                    else:
                        mother = mother[0].__bases__
                if not find:  # Not the good type
                    raise CheckTypeError(
                        "For "
                        + var_name
                        + " : "
                        + expect_type
                        + " expected, "
                        + type_value
                        + " given"
                    )


def check_min(var_name, value, type_value, Vmin):
    """Check if value is greater than the min of var_name
    Parameters
    ----------
    var_name : str
        The name of the property to set
    value : ?
        The value to check (int, float or ndarray)
    type_value : str
        The name of the actual type of the variable
    Vmin : float
        Value must be >Vmin
    Returns
    -------
    None
    Raises
    ------
    CheckMinError
        value is too small for var_nam
    """
    # Work only for number and matrix
    msg = "Check : You can't specify the min of variable of type " + type_value
    assert type_value in ["ndarray", "int", "long", "float", "int32"], msg
    # For number
    if (isinstance(value, int) or isinstance(value, float)) and value < Vmin:
        raise CheckMinError(
            var_name + " must be >= " + str(Vmin) + ", " + str(value) + " given"
        )
    # For non empty matrix
    if type_value == "ndarray" and value.size > 0 and value.min() < Vmin:
        raise CheckMinError(
            var_name
            + " must have its elements >= "
            + str(Vmin)
            + ", "
            + str(value)
            + " given"
        )


def check_max(var_name, value, type_value, Vmax):
    """Check if value is less than the max of var_name
    Parameters
    ----------
    var_name : str
        The name of the property to set
    value : ?
        The value to check (int, float or ndarray)
    type_value : str
        The name of the actual type of the variable
    Vmax : float
        Value must be <Vmax
    Returns
    -------
    None
    Raises
    ------
    CheckMaxError
        value is too large for var_nam
    """
    # Work only for number and matrix
    msg = "Check : You can't specify the max of variable of type " + type_value
    assert type_value in ["ndarray", "int", "long", "float", "int32"], msg
    # For number
    if (isinstance(value, int) or isinstance(value, float)) and value > Vmax:
        raise CheckMaxError(
            var_name + " must be <= " + str(Vmax) + ", " + str(value) + " given"
        )
    # For non empty matrix
    if type_value == "ndarray" and value.size > 0 and value.max() > Vmax:
        raise CheckMaxError(
            var_name
            + " must have its elements <= "
            + str(Vmax)
            + ", "
            + str(value)
            + " given"
        )


def raise_(ex):
    """Function to raise an exeption for the method import lambda"""
    raise ex


def check_dimensions(values, axes):
    """Check if field and axes have matching dimensions
    Parameters
    ----------
    values : ndarray
        values array
    axes : [Data]
        list of Data objects
    Returns
    -------
    None
    Raises
    ------
    CheckDimError
        Dimensions of field and axes do not match
    """
    if isinstance(values, list):
        values = array(values)
    else:
        values = values
    values_shape = values.shape
    axes_shape = []
    for axis in axes:
        if hasattr(axis, "values"):
            axes_shape.append(axis.values.size)
        else:
            try:
                axes_shape.append(int(axis.number))
            except:
                axes_shape.append(round((axis.final - axis.initial) / axis.step))

    if values_shape != tuple(axes_shape):
        for i, s in enumerate(axes_shape):
            if s == 1:
                values = expand_dims(values, axis=i)
        if values.shape != tuple(axes_shape):
            raise CheckDimError(
                "Dimensions of field ("
                + str(values_shape)
                + ") and axes ("
                + str(tuple(axes_shape))
                + ") do not match"
            )
    return values


class CheckDimError(Exception):
    """ """

    pass


class CheckError(Exception):
    """ """

    pass


class CheckMinError(CheckError):
    """ """

    pass


class CheckMaxError(CheckError):
    """ """

    pass


class CheckTypeError(CheckError):
    """ """

    pass


class InitUnKnowClassError(CheckError):
    """ """

    pass


class UnknowInitDictKeyError(CheckError):
    """ """

    pass


class NotADictError(CheckError):
    """ """

    pass



# -*- coding: utf-8 -*-


def freq_to_time(self):
    """Performs the Fourier Transform and stores the resulting field in a VectorField of DataTime objects.
    Parameters
    ----------
    self : VectorField
        a VectorField object
    Returns
    -------
    a VectorField object
    """

    # Dynamic import to avoid loop
    module = __import__("SciDataTool.Classes.VectorField", fromlist=["VectorField"])
    VectorField = getattr(module, "VectorField")

    comp_dict = dict()
    for key, comp in self.components.items():
        comp_dict[key] = comp.freq_to_time()

    return VectorField(name=self.name, symbol=self.symbol, components=comp_dict)



# -*- coding: utf-8 -*-
from numpy import (
    sum as np_sum,
    mean as np_mean,
    sqrt,
    trapz,
    take,
    min as np_min,
    max as np_max,
)
from SciDataTool.Functions.symmetries import rebuild_symmetries


def get_field(self, axes_list):
    """Returns the values of the field (with symmetries and sums).
    Parameters
    ----------
    self: Data
        a Data object
    axes_list: list
        a list of RequestedAxis objects
    Returns
    -------
    values: ndarray
        values of the field
    """

    values = self.values
    for axis_requested in axes_list:
        # Rebuild symmetries when needed
        axis_symmetries = self.axes[axis_requested.index].symmetries
        if (
            axis_requested.transform == "fft"
            and axis_requested.is_pattern
            or axis_requested.extension in ["sum", "rss", "mean", "rms", "integrate"]
            and axis_requested.is_pattern
        ):
            values = take(values, axis_requested.rebuild_indices, axis_requested.index)
        elif axis_requested.transform == "fft" and "antiperiod" in axis_symmetries:
            nper = axis_symmetries["antiperiod"]
            axis_symmetries["antiperiod"] = 2
            values = rebuild_symmetries(values, axis_requested.index, axis_symmetries)
            axis_symmetries["antiperiod"] = nper
        elif axis_requested.indices is not None:
            if (
                axis_requested.extension in ["sum", "rss", "mean", "rms", "integrate"]
                or max(axis_requested.indices) > values.shape[axis_requested.index]
            ):
                values = rebuild_symmetries(
                    values, axis_requested.index, axis_symmetries
                )
                self.axes[axis_requested.index].symmetries = dict()

        # # sum over sum axes
        # if axis_requested.extension == "sum":
        #     values = np_sum(values, axis=axis_requested.index, keepdims=True)
        # # root sum square over rss axes
        # elif axis_requested.extension == "rss":
        #     values = sqrt(np_sum(values ** 2, axis=axis_requested.index, keepdims=True))
        # # mean value over mean axes
        # elif axis_requested.extension == "mean":
        #     values = np_mean(values, axis=axis_requested.index, keepdims=True)
        # # RMS over rms axes
        # elif axis_requested.extension == "rms":
        #     values = sqrt(
        #         np_mean(values ** 2, axis=axis_requested.index, keepdims=True)
        #     )
        # # integration over integration axes
        # elif axis_requested.extension == "integrate":
        #     values = trapz(
        #         values, x=axis_requested.values, axis=axis_requested.index
        #     ) / (np_max(axis_requested.values) - np_min(axis_requested.values))
    return values



unit_dict = {
    "time": "s",
    "angle": "rad",
    "freqs": "Hz",
    "wavenumber": "",
    "phase": "",
    "z": "m",
    "radius": "m",
    "distance": "m",
    "loadcases": "",
    "eccentricity": "%",
    "frequency": "Hz",
    "revolution": "",
    "order": "",
    "cr_band": "Bark",
    "speed": "rpm",
}

norm_dict = {
    "elec_order": "Electrical order []",
    "mech_order": "Mechanical order []",
    "space_order": "Space order []",
    "distance": "Distance [m]",
    "angle_rotor": "Rotor mechanical angle [°]",
    "tooth_id": "Stator tooth number []",
    "speed": "Speed [rpm]",
    "Hz": "Frequency [Hz]",
    "Bark": "Critical band rate [Bark]",
    "time": "Time [s]",
}

axes_dict = {
    "freqs": "frequency",
    "cr_band": "critical band rate",
    "z": "axial direction",
}

fft_dict = {
    "time": "freqs",
    "angle": "wavenumber",
}

ifft_dict = {
    "freqs": "time",
    "wavenumber": "angle",
}

# Colors and linestyles
COLORS = [
    "tab:blue",
    "tab:orange",
    "tab:green",
    "tab:red",
    "tab:purple",
    "tab:brown",
    "tab:pink",
    "tab:gray",
    "tab:olive",
    "tab:cyan",
    "b",
    "g",
    "m",
    "yellow",
    "black",
    "lawngreen",
    "dodgerblue",
    "darkviolet",
    "deeppink",
]

LINESTYLES = [
    "solid",
    "dotted",
    "dashed",
    "dashdot",
    (0, (1, 10)),
    (0, (5, 10)),
    (0, (3, 5, 1, 5)),
    (0, (3, 10, 1, 10)),
    (0, (3, 5, 1, 5, 1, 5)),
    (0, (3, 1, 1, 1, 1, 1)),
]

## Paramater for Schematics plot
P_FONT_SIZE = 12  # Point Font size
SC_FONT_SIZE = 12  # Schematics Font size
TEXT_BOX = dict(  # Parameter of the text box
    boxstyle="round",
    ec=(0.0, 0.0, 0.0),
    fc=(1.0, 1.0, 1.0),
)
# Arrow parameters
ARROW_WIDTH = 2
ARROW_COLOR = "black"
# Schematics lines
SC_LINE_COLOR = "black"
SC_LINE_STYLE = "dotted"
SC_LINE_WIDTH = 1
# Main lines
MAIN_LINE_COLOR = "0.5"  # Gray
MAIN_LINE_STYLE = "dotted"
MAIN_LINE_WIDTH = 1



# -*- coding: utf-8 -*-
from SciDataTool.Functions.conversions import convert


def get_values(
    self,
    unit="SI",
    is_pattern=False,
    is_oneperiod=False,
    is_antiperiod=False,
    is_smallestperiod=False,
):
    """Returns the vector 'axis' taking symmetries into account.
    Parameters
    ----------
    self: DataPattern
        a DataPattern object
    unit: str
        requested unit
    is_oneperiod: bool
        return values on a single period
    is_antiperiod: bool
        return values on a semi period (only for antiperiodic signals)
    Returns
    -------
    Vector of axis values
    """
    values = self.values

    # Unit conversion
    if unit != "SI" and unit != self.unit:
        values = convert(values, self.unit, unit)

    # Rebuild pattern
    if is_smallestperiod or is_pattern:
        return values
    else:
        if self.values_whole is not None:
            return self.values_whole
        else:
            return values[self.rebuild_indices]



# -*- coding: utf-8 -*-


def has_period(self):
    """Indicates if a field has symmetries.
    Parameters
    ----------
    self: DataND
        a DataND object
    Returns
    -------
    Boolean
    """
    answer = False

    for axis in self.axes:

        if axis.symmetries != {}:
            answer = True

    return answer



# -*- coding: utf-8 -*-
from numpy import (
    abs as np_abs,
    sum as np_sum,
    mean as np_mean,
    sqrt,
    trapz,
    squeeze,
    min as np_min,
    max as np_max,
    apply_along_axis,
)

from SciDataTool.Functions import NormError, UnitError
from SciDataTool.Functions.conversions import convert as convert_unit, to_dB, to_dBA


def convert(self, values, unit, is_norm, is_squeeze, is_magnitude, axes_list):
    """Returns the values of the field transformed or converted.
    Parameters
    ----------
    self: Data
        a Data object
    values: ndarray
        array of the field
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    Returns
    -------
    values: ndarray
        values of the field
    """

    # Take magnitude before summing
    if is_magnitude:
        values = np_abs(values)

    # Apply sums, means, etc
    for axis_requested in axes_list:
        # sum over sum axes
        if axis_requested.extension == "sum":
            values = np_sum(values, axis=axis_requested.index, keepdims=True)
        # root sum square over rss axes
        elif axis_requested.extension == "rss":
            values = sqrt(np_sum(values ** 2, axis=axis_requested.index, keepdims=True))
        # mean value over mean axes
        elif axis_requested.extension == "mean":
            values = np_mean(values, axis=axis_requested.index, keepdims=True)
        # RMS over rms axes
        elif axis_requested.extension == "rms":
            values = sqrt(
                np_mean(values ** 2, axis=axis_requested.index, keepdims=True)
            )
        # integration over integration axes
        elif axis_requested.extension == "integrate":
            values = trapz(
                values, x=axis_requested.values, axis=axis_requested.index
            ) / (np_max(axis_requested.values) - np_min(axis_requested.values))

    if is_squeeze:
        values = squeeze(values)

    if unit == self.unit or unit == "SI":
        if is_norm:
            try:
                values = values / self.normalizations.get("ref")
            except:
                raise NormError(
                    "ERROR: Reference value not specified for normalization"
                )
    elif unit == "dB":
        ref_value = 1.0
        if "ref" in self.normalizations.keys():
            ref_value *= self.normalizations.get("ref")
        values = to_dB(np_abs(values), self.unit, ref_value)
    elif unit == "dBA":
        ref_value = 1.0
        if "ref" in self.normalizations.keys():
            ref_value *= self.normalizations.get("ref")
        for axis in axes_list:
            is_match = False
            if axis.name == "freqs" or axis.corr_name == "freqs":
                index = axis.index
                values = apply_along_axis(
                    to_dBA, index, values, axis.values, self.unit, ref_value
                )
                is_match = True
            elif axis.name == "frequency":
                index = axis.index
                values = apply_along_axis(
                    to_dBA, index, values, axis.values, self.unit, ref_value
                )
                is_match = True
        if not is_match:
            raise UnitError(
                "ERROR: dBA conversion only available for fft with frequencies"
            )
    elif unit in self.normalizations:
        values = values / self.normalizations.get(unit)
    else:
        values = convert_unit(values, self.unit, unit)

    return values



# -*- coding: utf-8 -*-
from SciDataTool.Functions import AxisError


def get_length(self, is_pattern=False):
    """Returns the length of the axis taking symmetries into account.
    Parameters
    ----------
    self: DataPattern
        a DataPattern object
    is_pattern: bool
        return length of pattern
    Returns
    -------
    Length of axis
    """

    if is_pattern:
        return len(self.values)
    else:
        return len(self.rebuild_indices)



# -*- coding: utf-8 -*-

from SciDataTool.Functions.Plot.plot_2D import plot_2D
from SciDataTool.Functions.Plot import (
    unit_dict,
    norm_dict,
    axes_dict,
    COLORS,
    LINESTYLES,
)
from numpy import (
    squeeze,
    split,
    array,
    where,
    unique,
    max as np_max,
    array2string,
    insert,
)


def plot_2D_Data(
    self,
    *arg_list,
    axis_data=None,
    is_norm=False,
    unit="SI",
    data_list=[],
    legend_list=[],
    color_list=None,
    curve_colors=None,
    phase_colors=None,
    linestyles=None,
    linewidth_list=[2],
    save_path=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_disp_title=True,
    is_grid=True,
    is_auto_ticks=True,
    is_auto_range=True,
    fig=None,
    ax=None,
    barwidth=100,
    type_plot=None,
    fund_harm_dict=None,
    is_show_fig=None,
    win_title=None,
    thresh=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
):
    """Plots a field as a function of time

    Parameters
    ----------
    data : Data
        a Data object
    *arg_list : list of str
        arguments to specify which axes to plot
    is_norm : bool
        boolean indicating if the field must be normalized
    unit : str
        unit in which to plot the field
    data_list : list
        list of Data objects to compare
    legend_list : list
        list of legends to use for each Data object (including reference one) instead of data.name
    color_list : list
        list of colors to use for each Data object
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_disp_title : bool
        boolean indicating if the title must be displayed
    is_grid : bool
        boolean indicating if the grid must be displayed
    is_auto_ticks : bool
        in fft, adjust ticks to freqs (deactivate if too close)
    is_auto_range : bool
        in fft, display up to 1% of max
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    barwidth : float
        barwidth scaling factor, only if type_plot = "bargraph"
    type_plot : str
        type of 2D graph : "curve", "bargraph", "barchart" or "quiver"
    fund_harm_dict : dict
        Dict containing axis name as key and frequency/order/wavenumber of fundamental harmonic as value to display fundamental harmonic in red in the fft
    is_show_fig : bool
        True to show figure after plot
    win_title : str
        Title of the plot window
    thresh : float
        threshold for automatic fft ticks
    """

    # Extract arg_list it the function called from another script with *arg_list
    if len(arg_list) == 1 and type(arg_list[0]) == tuple:
        arg_list = arg_list[0]

    if color_list == []:
        color_list = COLORS

    # Set unit
    if unit == "SI":
        unit = self.unit

    # Detect if is fft, build first title part and ylabel
    is_fft = False
    if any("wavenumber" in s for s in arg_list) or any("freqs" in s for s in arg_list):
        is_fft = True
        if "dB" in unit:
            unit_str = (
                "[" + unit + " re. " + str(self.normalizations["ref"]) + self.unit + "]"
            )
        else:
            unit_str = "[" + unit + "]"
        if self.symbol == "Magnitude":
            ylabel = "Magnitude " + r"$[" + unit + "]$"
        else:
            ylabel = r"$|\widehat{" + self.symbol + "}|$ " + unit_str
        if data_list == []:
            title1 = "FFT of " + self.name.lower() + " "
        else:
            title1 = "Comparison of " + self.name.lower() + " FFT "
    else:
        if data_list == []:
            # title = data.name.capitalize() + " for " + ', '.join(arg_list_str)
            title1 = self.name.capitalize() + " "
        else:
            # title = data.name.capitalize() + " for " + ', '.join(arg_list_str)
            title1 = "Comparison of " + self.name.lower() + " "
        if is_norm:
            ylabel = (
                r"$\frac{" + self.symbol + "}{" + self.symbol + "_0}\, [" + unit + "]$"
            )
        else:
            if self.symbol == "Magnitude":
                ylabel = "Magnitude " + r"$[" + unit + "]$"
            else:
                ylabel = r"$" + self.symbol + "\, [" + unit + "]$"

    # Extract field and axes
    Xdatas = []
    Ydatas = []
    data_list2 = [self] + data_list
    for i, d in enumerate(data_list2):
        if is_fft:
            result = d.get_magnitude_along(
                arg_list, axis_data=axis_data, unit=unit, is_norm=is_norm
            )
            if i == 0:
                axes_list = result.pop("axes_list")
                axes_dict_other = result.pop("axes_dict_other")
                result_0 = result
            Ydatas.append(result.pop(d.symbol))
            Xdatas.append(result[list(result)[0]])
        else:
            result = d.get_along(
                arg_list, axis_data=axis_data, unit=unit, is_norm=is_norm
            )
            if i == 0:
                axes_list = result.pop("axes_list")
                axes_dict_other = result.pop("axes_dict_other")
                result_0 = result
            Ydatas.append(result.pop(d.symbol))
            Xdatas.append(result[list(result)[0]])

    # Find main axis as the axis with the most values
    # main_axis = axes_list[0]  # max(axes_list, key=lambda x: x.values.size)

    # Build xlabel and title parts 2 and 3
    title2 = ""
    title3 = " for "
    for axis in axes_list:
        # Title part 2
        if axis.unit in norm_dict:
            name = norm_dict[axis.unit].split(" [")[0]
        elif axis.name in axes_dict:
            name = axes_dict[axis.name]
        else:
            name = axis.name
        if (
            axis.extension
            in [
                "whole",
                "interval",
                "oneperiod",
                "antiperiod",
                "smallestperiod",
                "axis_data",
            ]
            and len(axis.values) > 1
        ):
            title2 = "over " + name.lower()
            if axis.unit == "SI":
                unit = unit_dict[axis.name]
                xlabel = name.capitalize() + " [" + unit + "]"
                main_axis_name = name
            elif axis.unit in norm_dict:
                xlabel = norm_dict[axis.unit]
                if axis.unit == "Hz":
                    main_axis_name = "frequency"
                else:
                    main_axis_name = axis.unit
            else:
                unit = axis.unit
                xlabel = name.capitalize() + " [" + unit + "]"
                main_axis_name = name
            if (
                axis.name == "angle"
                and axis.unit == "°"
                and round(np_max(axis.values) / 6) % 5 == 0
            ):
                xticks = [i * round(np_max(axis.values) / 6) for i in range(7)]
            else:
                xticks = None
        else:
            if axis.unit == "SI":
                unit = unit_dict[axis.name]
            elif axis.unit in norm_dict:
                unit = norm_dict[axis.unit]
            else:
                unit = axis.unit

            axis_str = array2string(
                result_0[axis.name], formatter={"float_kind": "{:.3g}".format}
            ).replace(" ", ", ")
            if len(result_0[axis.name]) == 1:
                axis_str = axis_str.strip("[]")

            title3 += name + "=" + axis_str.rstrip(", ") + " [" + unit + "], "

    # Title part 4 containing axes that are here but not involved in requested axes
    title4 = ""
    for axis_name in axes_dict_other:
        title4 += (
            axis_name
            + "="
            + array2string(
                axes_dict_other[axis_name][0],
                formatter={"float_kind": "{:.3g}".format},
            ).replace(" ", ", ")
            + " ["
            + axes_dict_other[axis_name][1]
            + "], "
        )

    if title3 == " for " and title4 == "":
        title3 = ""

    # Concatenate all title parts
    title = title1 + title2 + title3 + title4

    # Remove last coma due to title3 or title4
    title = title.rstrip(", ")

    # Remove dimless and quotes
    title = title.replace("[]", "")
    title = title.replace("'", "")

    # Detect how many curves are overlaid, build legend and color lists
    if legend_list == [] and data_list != []:
        legend_list = ["[" + d.name + "] " for d in data_list2]
    elif legend_list == []:
        legend_list = ["" for d in data_list2]
    else:
        legend_list = ["[" + leg + "] " for leg in legend_list]
    legends = []
    # Prepare colors
    if curve_colors is None:
        curve_colors = COLORS
    linestyle_list = linestyles
    for i, d in enumerate(data_list2):
        is_overlay = False
        for axis in axes_list:
            if axis.extension == "list":
                is_overlay = True
                if phase_colors is None:
                    phase_colors = COLORS
                    if linestyles is None:
                        linestyles = ["dashed"]
                n_curves = len(axis.values)
                if axis.unit == "SI":
                    unit = unit_dict[axis.name]
                elif axis.unit in norm_dict:
                    unit = norm_dict[axis.unit]
                else:
                    unit = axis.unit
                if len(d.axes[axis.index].get_values()) > 1:
                    legends += [
                        legend_list[i]
                        + axis.name
                        + "="
                        + axis.values.tolist()[j]
                        + " "
                        + unit
                        if isinstance(axis.values.tolist()[j], str)
                        else legend_list[i]
                        + axis.name
                        + "="
                        + "%.3g" % axis.values.tolist()[j]
                        + " "
                        + unit
                        for j in range(n_curves)
                    ]
                else:
                    legends += [legend_list[i]]

        if not is_overlay:
            legends += [legend_list[i]]

    # Split Ydatas if the plot overlays several curves
    if is_overlay:
        Ydata = []
        for d in Ydatas:
            if d.ndim != 1:
                axis_index = where(array(d.shape) == n_curves)[0]
                if axis_index.size > 1:
                    print("WARNING, several axes with same dimensions")
                Ydata += split(d, n_curves, axis=axis_index[0])
            else:
                Ydata += [d]
        Ydatas = [squeeze(d) for d in Ydata]
        Xdata = []
        for i in range(len(data_list2)):
            Xdata += [Xdatas[i] for x in range(n_curves)]
        Xdatas = Xdata

    # Call generic plot function
    if is_fft:

        if thresh is None:
            if self.normalizations is not None and "ref" in self.normalizations.keys():
                thresh = self.normalizations["ref"]
            else:
                thresh = 0.02

        freqs = Xdatas[0]
        indices = [
            ind
            for ind, y in enumerate(Ydatas[0])
            if abs(y) > abs(thresh * np_max(Ydatas[0]))
        ]
        xticks = unique(insert(0, 0, freqs[indices]))

        if is_auto_range:
            if len(xticks) > 0:
                x_min = -xticks[-1] * 0.1
                x_max = xticks[-1] * 1.1

        if not is_auto_ticks:
            xticks = None

        # Force bargraph for fft if type_graph not specified
        if type_plot is None:
            type_plot = "bargraph"

        # Option to draw fundamental harmonic in red
        if not fund_harm_dict:
            fund_harm = None
        else:
            # Activate the option only if main axis is in dict and only one Data is plotted
            if main_axis_name in fund_harm_dict and len(Ydatas) == 1:
                fund_harm = fund_harm_dict[main_axis_name]
            else:
                # Deactivate the option
                fund_harm = None

        plot_2D(
            Xdatas,
            Ydatas,
            legend_list=legends,
            color_list=color_list,
            linestyle_list=linestyle_list,
            linewidth_list=linewidth_list,
            fig=fig,
            ax=ax,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            type_plot=type_plot,
            x_min=x_min,
            x_max=x_max,
            y_min=y_min,
            y_max=y_max,
            is_logscale_x=is_logscale_x,
            is_logscale_y=is_logscale_y,
            is_disp_title=is_disp_title,
            is_grid=is_grid,
            xticks=xticks,
            save_path=save_path,
            barwidth=barwidth,
            fund_harm=fund_harm,
            is_show_fig=is_show_fig,
            win_title=win_title,
            font_name=font_name,
            font_size_title=font_size_title,
            font_size_label=font_size_label,
            font_size_legend=font_size_legend,
        )

    else:

        # Force curve plot if type_plot not specified
        if type_plot is None:
            type_plot = "curve"

        plot_2D(
            Xdatas,
            Ydatas,
            legend_list=legends,
            color_list=color_list,
            fig=fig,
            ax=ax,
            title=title,
            xlabel=xlabel,
            ylabel=ylabel,
            type_plot=type_plot,
            x_min=x_min,
            x_max=x_max,
            y_min=y_min,
            y_max=y_max,
            is_logscale_x=is_logscale_x,
            is_logscale_y=is_logscale_y,
            is_disp_title=is_disp_title,
            is_grid=is_grid,
            xticks=xticks,
            barwidth=barwidth,
            linestyle_list=linestyle_list,
            linewidth_list=linewidth_list,
            save_path=save_path,
            is_show_fig=is_show_fig,
            win_title=win_title,
            font_name=font_name,
            font_size_title=font_size_title,
            font_size_label=font_size_label,
            font_size_legend=font_size_legend,
        )



def plot_3D_Data(
    self,
    *arg_list,
    axis_data=None,
    is_norm=False,
    unit="SI",
    component_list=None,
    save_path=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    z_min=None,
    z_max=None,
    z_range=None,
    is_auto_ticks=True,
    is_auto_range=False,
    is_2D_view=False,
    is_same_size=False,
    N_stem=100,
    fig=None,
    ax=None,
    is_show_fig=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_logscale_z=False,
    thresh=None,
    is_switch_axes=False,
    colormap="RdBu_r",
    win_title=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
):
    """Plots a field as a function of time

    Parameters
    ----------
    self : Output
        an Output object
    Data_str : str
        name of the Data Object to plot (e.g. "mag.Br")
    *arg_list : list of str
        arguments to specify which axes to plot
    is_norm : bool
        boolean indicating if the field must be normalized
    unit : str
        unit in which to plot the field
    save_path : str
        full path including folder, name and extension of the file to save if save_path is not None
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    z_min : float
        minimum value for the z-axis
    z_max : float
        maximum value for the z-axis
    is_auto_ticks : bool
        in fft, adjust ticks to freqs (deactivate if too close)
    is_auto_range : bool
        in fft, display up to 1% of max
    is_2D_view : bool
        True to plot Data in xy plane and put z as colormap
    is_same_size : bool
        True to have all color blocks with same size in 2D view
    N_stem : int
        number of harmonics to plot (only for stem plots)
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    is_show_fig : bool
        True to show figure after plot
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_logscale_z : bool
        boolean indicating if the z-axis must be set in logarithmic scale
    thresh : float
        threshold for automatic fft ticks
    is_switch_axes : bool
        to switch x and y axes
    """

    # Call the plot on each component
    if component_list is None:  # default: extract all components
        component_list = self.components.keys()
    for i, comp in enumerate(component_list):

        if save_path is not None and len(component_list) > 1:
            save_path_comp = (
                save_path.split(".")[0] + "_" + comp + "." + save_path.split(".")[1]
            )
        else:
            save_path_comp = save_path

        self.components[comp].plot_3D_Data(
            arg_list,
            axis_data=axis_data,
            is_norm=is_norm,
            unit=unit,
            save_path=save_path_comp,
            x_min=x_min,
            x_max=x_max,
            y_min=y_min,
            y_max=y_max,
            z_min=z_min,
            z_max=z_max,
            colormap=colormap,
            is_auto_ticks=is_auto_ticks,
            is_auto_range=is_auto_range,
            is_2D_view=is_2D_view,
            is_same_size=is_same_size,
            N_stem=N_stem,
            fig=fig,
            ax=ax,
            is_show_fig=is_show_fig,
            is_logscale_x=is_logscale_x,
            is_logscale_y=is_logscale_y,
            is_logscale_z=is_logscale_z,
            thresh=thresh,
            is_switch_axes=is_switch_axes,
            win_title=win_title,
            font_name=font_name,
            font_size_title=font_size_title,
            font_size_label=font_size_label,
            font_size_legend=font_size_legend,
        )



# -*- coding: utf-8 -*-
from SciDataTool.Functions import AxisError


def get_length(self, is_oneperiod=False, is_antiperiod=False):
    """Returns the length of the axis taking symmetries into account.
    Parameters
    ----------
    self: DataLinspace
        a DataLinspace object
    is_oneperiod: bool
        return values on a single period
    is_antiperiod: bool
        return values on a semi period (only for antiperiodic signals)
    Returns
    -------
    Length of axis
    """

    if self.number is None:
        N = round((self.final - self.initial + self.step) / self.step)
    else:
        N = self.number

    # Rebuild symmetries
    if is_antiperiod:
        if "antiperiod" in self.symmetries:
            return N
        else:
            raise AxisError("ERROR: axis has no antiperiodicity")
    elif is_oneperiod:
        if "antiperiod" in self.symmetries:
            return N * 2
        elif "period" in self.symmetries:
            return N
        else:
            return N
    else:
        if "antiperiod" in self.symmetries:
            return N * self.symmetries["antiperiod"]
        elif "period" in self.symmetries:
            return N * self.symmetries["period"]
        else:
            return N



# -*- coding: utf-8 -*-
from SciDataTool.Functions import AxisError, axes_dict, rev_axes_dict

from numpy import inf


def comp_axes(self, axes_list):
    """Completes the RequestedAxis objects in axes_list.
    Parameters
    ----------
    self: Data
        a Data object
    axes_list: list
        a list of RequestedAxis objects
    Returns
    -------
    list of RequestedAxis objects + transformations
    """

    transforms = []
    # Check if the requested axis is defined in the Data object
    for axis_requested in axes_list[:]:
        axis_name = axis_requested.name
        for index, axis in enumerate(self.axes):
            if axis.name == axis_name:
                axis_requested.index = index
                axis_requested.corr_name = axis_name
                axis_requested.corr_unit = axis.unit
                if axis_name in rev_axes_dict.keys():
                    axis_requested.transform = "fft_axis"
                    transforms.append("fft")
        if axis_requested.index is None:
            # Check if requested axis is in correspondance dicts
            if axis_name in axes_dict.keys():
                for index, axis in enumerate(self.axes):
                    if axis.name == axes_dict[axis_name][0]:
                        axis_requested.corr_name = axes_dict[axis_name][0]
                        axis_requested.corr_unit = axes_dict[axis_name][2]
                        axis_requested.operation = (
                            axes_dict[axis_name][0] + "_to_" + axis_name
                        )
                        axis_requested.transform = axes_dict[axis_name][1]
                        transforms.append(axes_dict[axis_name][1])
                        axis_requested.index = index
                if axis_requested.index is None:
                    # Axis does not exist and is ignored
                    axes_list.remove(axis_requested)
            elif axis_name in rev_axes_dict.keys():
                for index, axis in enumerate(self.axes):
                    if axis.name == rev_axes_dict[axis_name][0]:
                        axis_requested.corr_name = rev_axes_dict[axis_name][0]
                        axis_requested.corr_unit = rev_axes_dict[axis_name][2]
                        axis_requested.operation = (
                            rev_axes_dict[axis_name][0] + "_to_" + axis_name
                        )
                        axis_requested.transform = rev_axes_dict[axis_name][1]
                        transforms.append(rev_axes_dict[axis_name][1])
                        axis_requested.index = index
                if axis_requested.index is None:
                    # Axis does not exist and is ignored
                    axes_list.remove(axis_requested)
            else:
                # Axis does not exist and is ignored
                axes_list.remove(axis_requested)
    # Extract the requested axes (symmetries + unit)
    for axis_requested in axes_list:
        axis_requested.get_axis(self.axes[axis_requested.index], is_real=self.is_real)
    return axes_list, transforms



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/Data.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//Data
"""

from os import linesep
from sys import getsizeof
from ._check import check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from ._frozen import FrozenClass

from ._check import InitUnKnowClassError


class Data(FrozenClass):
    """Abstract class for all kinds of data"""

    VERSION = 1

    # save and copy methods are available in all object
    save = save
    copy = copy

    def __init__(
        self,
        symbol="",
        name="",
        unit="",
        normalizations=-1,
        init_dict=None,
        init_str=None,
    ):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionnary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "symbol" in list(init_dict.keys()):
                symbol = init_dict["symbol"]
            if "name" in list(init_dict.keys()):
                name = init_dict["name"]
            if "unit" in list(init_dict.keys()):
                unit = init_dict["unit"]
            if "normalizations" in list(init_dict.keys()):
                normalizations = init_dict["normalizations"]
        # Set the properties (value check and convertion are done in setter)
        self.parent = None
        self.symbol = symbol
        self.name = name
        self.unit = unit
        self.normalizations = normalizations

        # The class is frozen, for now it's impossible to add new properties
        self._freeze()

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        Data_str = ""
        if self.parent is None:
            Data_str += "parent = None " + linesep
        else:
            Data_str += "parent = " + str(type(self.parent)) + " object" + linesep
        Data_str += 'symbol = "' + str(self.symbol) + '"' + linesep
        Data_str += 'name = "' + str(self.name) + '"' + linesep
        Data_str += 'unit = "' + str(self.unit) + '"' + linesep
        Data_str += "normalizations = " + str(self.normalizations) + linesep
        return Data_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False
        if other.symbol != self.symbol:
            return False
        if other.name != self.name:
            return False
        if other.unit != self.unit:
            return False
        if other.normalizations != self.normalizations:
            return False
        return True

    def compare(self, other, name="self"):
        """Compare two objects and return list of differences"""

        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()
        if other._symbol != self._symbol:
            diff_list.append(name + ".symbol")
        if other._name != self._name:
            diff_list.append(name + ".name")
        if other._unit != self._unit:
            diff_list.append(name + ".unit")
        if other._normalizations != self._normalizations:
            diff_list.append(name + ".normalizations")
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object
        S += getsizeof(self.symbol)
        S += getsizeof(self.name)
        S += getsizeof(self.unit)
        if self.normalizations is not None:
            for key, value in self.normalizations.items():
                S += getsizeof(value) + getsizeof(key)
        return S

    def as_dict(self, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        Data_dict = dict()
        Data_dict["symbol"] = self.symbol
        Data_dict["name"] = self.name
        Data_dict["unit"] = self.unit
        Data_dict["normalizations"] = (
            self.normalizations.copy() if self.normalizations is not None else None
        )
        # The class name is added to the dict for deserialisation purpose
        Data_dict["__class__"] = "Data"
        return Data_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        self.symbol = None
        self.name = None
        self.unit = None
        self.normalizations = None

    def _get_symbol(self):
        """getter of symbol"""
        return self._symbol

    def _set_symbol(self, value):
        """setter of symbol"""
        check_var("symbol", value, "str")
        self._symbol = value

    symbol = property(
        fget=_get_symbol,
        fset=_set_symbol,
        doc=u"""Symbol of the variable (in latex syntax)

        :Type: str
        """,
    )

    def _get_name(self):
        """getter of name"""
        return self._name

    def _set_name(self, value):
        """setter of name"""
        check_var("name", value, "str")
        self._name = value

    name = property(
        fget=_get_name,
        fset=_set_name,
        doc=u"""Name of the physical quantity (to be used in plots)

        :Type: str
        """,
    )

    def _get_unit(self):
        """getter of unit"""
        return self._unit

    def _set_unit(self, value):
        """setter of unit"""
        check_var("unit", value, "str")
        self._unit = value

    unit = property(
        fget=_get_unit,
        fset=_set_unit,
        doc=u"""Unit of the physical quantity (to be used in plots)

        :Type: str
        """,
    )

    def _get_normalizations(self):
        """getter of normalizations"""
        return self._normalizations

    def _set_normalizations(self, value):
        """setter of normalizations"""
        if type(value) is int and value == -1:
            value = dict()
        check_var("normalizations", value, "dict")
        self._normalizations = value

    normalizations = property(
        fget=_get_normalizations,
        fset=_set_normalizations,
        doc=u"""Normalizations available for the field and its axes

        :Type: dict
        """,
    )



# -*- coding: utf-8 -*-
from SciDataTool.Functions.conversions import convert
from SciDataTool.Functions.interpolations import get_common_base
from SciDataTool.Functions.symmetries import rebuild_symmetries_axis
from SciDataTool.Classes.DataPattern import DataPattern
from SciDataTool.Functions import AxisError
from numpy import array, ndarray
from importlib import import_module


def get_axis(self, axis, is_real):
    """Computes the vector 'axis' in the unit required, using conversions and symmetries if needed.
    Parameters
    ----------
    self: RequestedAxis
        a RequestedAxis object
    axis: Axis
        an Axis object
    """
    if self.operation is not None:
        module = import_module("SciDataTool.Functions.conversions")
        func = getattr(module, self.operation)  # Conversion function
    if isinstance(axis, DataPattern):
        self.is_pattern = True
        self.rebuild_indices = axis.rebuild_indices
        self.is_step = axis.is_step
    is_components = getattr(axis, "is_components", False)
    if is_components:
        values = axis.get_values()
        if not self.extension in ["sum", "rss", "mean", "rms", "integrate"]:
            self.extension = "list"
        if self.indices is not None:
            self.values = values[self.indices]
        else:
            self.values = values
    else:
        if self.extension == "pattern":
            if not self.is_pattern:
                raise AxisError(
                    "ERROR: [pattern] cannot be called with non DataPattern axis"
                )
            else:
                is_smallestperiod = True
                is_oneperiod = False
                is_antiperiod = False
                self.extension = "smallestperiod"
        elif self.extension == "smallestperiod":
            if isinstance(axis, DataPattern):
                raise AxisError(
                    "ERROR: [smallestperiod] cannot be called with DataPattern axis"
                )
            else:
                is_smallestperiod = True
                is_oneperiod = False
                is_antiperiod = False
        elif self.extension == "antiperiod":
            if isinstance(axis, DataPattern):
                raise AxisError(
                    "ERROR: [antiperiod] cannot be called with DataPattern axis"
                )
            else:
                is_smallestperiod = False
                is_oneperiod = False
                is_antiperiod = True
        elif self.extension == "oneperiod" or self.transform == "fft":
            if isinstance(axis, DataPattern):
                raise AxisError(
                    "ERROR: [oneperiod] cannot be called with DataPattern axis"
                )
            else:
                is_smallestperiod = False
                is_oneperiod = True
                is_antiperiod = False
        elif self.extension in ["sum", "rss", "mean", "rms", "integrate"]:
            is_smallestperiod = False
            is_oneperiod = False
            is_antiperiod = False
        # Ignore symmetries if fft axis
        elif self.name == "freqs" or self.name == "wavenumber":
            is_smallestperiod = True
            is_oneperiod = False
            is_antiperiod = False
        else:
            if self.input_data is not None and not self.is_step:
                # Check if symmetries need to be reconstructed to match input_data
                if self.operation is not None:
                    axis_values = func(
                        axis.get_values(
                            is_smallestperiod=True,
                        ),
                        is_real=is_real,
                    )
                else:
                    axis_values = axis.get_values(
                        is_smallestperiod=True,
                    )
                if min(self.input_data) >= min(axis_values) and max(
                    self.input_data
                ) <= max(axis_values):
                    is_smallestperiod = True
                    is_oneperiod = False
                    is_antiperiod = False
                else:
                    if self.operation is not None:
                        axis_values = func(
                            axis.get_values(
                                is_oneperiod=True,
                            ),
                            is_real=is_real,
                        )
                    else:
                        axis_values = axis.get_values(
                            is_oneperiod=True,
                        )
                    if min(self.input_data) >= min(axis_values) and max(
                        self.input_data
                    ) <= max(axis_values):
                        is_smallestperiod = False
                        is_oneperiod = True
                        is_antiperiod = False
                        self.extension = "oneperiod"
                    else:
                        is_smallestperiod = False
                        is_oneperiod = False
                        is_antiperiod = False
                        if not self.is_pattern:
                            self.extension = "interval"
            elif self.transform == "ifft":  # Ignore symmetries in ifft case
                is_smallestperiod = True
                is_oneperiod = False
                is_antiperiod = False
            else:
                is_smallestperiod = False
                is_oneperiod = False
                is_antiperiod = False
        # Get original values of the axis
        if self.operation is not None:
            values = array(
                func(
                    axis.get_values(
                        is_oneperiod=is_oneperiod,
                        is_antiperiod=is_antiperiod,
                        is_smallestperiod=is_smallestperiod,
                    ),
                    is_real=is_real,
                )
            )
            # Store original values
            self.corr_values = array(
                axis.get_values(
                    is_oneperiod=is_oneperiod,
                    is_antiperiod=is_antiperiod,
                    is_smallestperiod=is_smallestperiod,
                )
            )
        else:
            values = array(
                axis.get_values(
                    is_oneperiod=is_oneperiod,
                    is_antiperiod=is_antiperiod,
                    is_smallestperiod=is_smallestperiod,
                )
            )
        # Unit conversions and normalizations
        unit = self.unit
        if unit == self.corr_unit or unit == "SI":
            pass
        elif unit in axis.normalizations:
            if axis.normalizations.get(unit) == "indices":
                values = array([i for i in range(len(values))])
            elif isinstance(axis.normalizations.get(unit), ndarray):
                values = axis.normalizations.get(unit)
            else:
                values = values / axis.normalizations.get(unit)
        else:
            values = convert(values, self.corr_unit, unit)
        # Rebuild symmetries in fft case
        if self.transform == "fft":
            if "period" in axis.symmetries:
                if axis.name != "time":
                    values = values * axis.symmetries["period"]
            elif "antiperiod" in axis.symmetries:
                if axis.name != "time":
                    values = values * axis.symmetries["antiperiod"] / 2
        # Rebuild symmetries in ifft case
        if self.transform == "ifft":
            # if "antiperiod" in axis.symmetries:
            #     axis.symmetries["antiperiod"] = int(axis.symmetries["antiperiod"]/2)
            if (
                self.extension != "smallestperiod"
                and self.extension != "oneperiod"
                and self.extension != "antiperiod"
            ):
                values = rebuild_symmetries_axis(values, axis.symmetries)
            # if "period" in axis.symmetries:
            #     if axis.name != "freqs":
            #         values = values * axis.symmetries["period"]
            # elif "antiperiod" in axis.symmetries:
            #     if axis.name != "freqs":
            #         values = values * axis.symmetries["antiperiod"] / 2
        # Interpolate axis with input data
        if self.input_data is None:
            self.values = values
        else:
            # if self.is_step:
            #     values = values[axis.rebuild_indices]
            if len(self.input_data) == 2 and self.extension != "axis_data":
                indices = [
                    i
                    for i, x in enumerate(values)
                    if x >= self.input_data[0] and x <= self.input_data[-1]
                ]
                if self.indices is None:
                    self.indices = indices
                else:
                    indices_new = []
                    for i in self.indices:
                        if i in indices:
                            indices_new.append(i)
                    self.indices = indices_new
                self.input_data = None
            else:
                self.values = values
        if self.indices is not None:
            self.values = values[self.indices]
            if self.extension in ["sum", "rss", "mean", "rms", "integrate"]:
                self.indices = None



# -*- coding: utf-8 -*-
from SciDataTool.Classes._check import check_dimensions, check_var
from numpy import squeeze, array


def _set_values(self, value):
    """setter of values"""
    if type(value) is int and value == -1:
        value = array([])
    elif type(value) is list:
        try:
            value = array(value)
        except:
            pass
    check_var("values", value, "ndarray")

    # Check dimensions
    value = squeeze(value)
    value = check_dimensions(value, self.axes)
    self._values = value



import matplotlib.pyplot as plt
import numpy as np
import imageio
from SciDataTool.Methods.DataND.plot_2D_Data import plot_2D_Data


def plot_2D_Data_Animated(
    self, animated_axis, *param_list, nb_frames=50, fps=10, **param_dict
):
    """Gen

    Parameters
    ----------
    animated_axis : str
        The field will be animated along this axis
    nb_frames : int
        number of frames used to build the gif
    fps: int
        frames displayed per second
    """

    # The list of images used to build the gif
    images = list()
    if "freqs" in param_list or "wavenumber" in param_list:
        result = self.get_magnitude_along(animated_axis, *param_list)
    else:
        result = self.get_along(animated_axis, *param_list)

    value_max = np.max(result[animated_axis])
    value_min = np.min(result[animated_axis])
    variation_step = (value_max - value_min) / nb_frames

    y_max = np.max(result[self.symbol])
    y_min = np.min(result[self.symbol])
    marge = (
        y_max - y_min
    ) * 0.05  # 5% of the height of plot to add to the border top/bottom of gif
    param_dict["y_min"] = np.min(result[self.symbol]) - abs(marge)
    param_dict["y_max"] = np.max(result[self.symbol]) + abs(marge)
    param_dict["is_show_fig"] = False
    # Getting the name of the gif
    save_path = param_dict["save_path"].replace(".png", ".gif")
    param_dict["save_path"] = None
    while value_min < value_max:
        # plotting image
        self.plot_2D_Data(
            *param_list, animated_axis + "=" + str(value_min), **param_dict
        )
        # Getting the figure generated with plot_2D_DATA
        fig = plt.gcf()
        fig.canvas.draw()
        image = np.frombuffer(fig.canvas.tostring_rgb(), dtype="uint8")
        image = image.reshape(fig.canvas.get_width_height()[::-1] + (3,))
        images.append(image)
        plt.close(fig)
        value_min += variation_step
    # Creating the gif
    imageio.mimsave(save_path, images, format="GIF-PIL", fps=fps)



# -*- coding: utf-8 -*-
from numpy import mean, hanning, linspace, where, isclose, apply_along_axis
from numpy.fft import (
    fft,
    fftshift,
    ifft,
    ifftshift,
    rfftn,
    irfftn,
    fftn,
    ifftn,
    rfftfreq,
)
from numpy import (
    array,
    pi,
    zeros,
    exp,
    ceil,
    iscomplex,
    concatenate,
    conjugate,
    flip,
    isin,
    append,
    take,
    insert,
    delete,
    abs as np_abs,
    angle as np_angle,
    allclose,
    real,
)

from SciDataTool.Functions.nudft_functions import is_uniform, nudftn, inudftn


def comp_fft_freqs(time, is_time, is_real):
    """Computes the frequency/wavenumber vector from the time/space vector
    Parameters
    ----------
    time: array
        Time or space vector
    is_time: bool
        Boolean indicating if we input time or space
    Returns
    -------
    Frequency/wavenumber vector
    """
    N_tot = time.size  # Number of samples
    if N_tot == 1:
        freqs = [0]
    else:
        # zero-padding
        # N_tot = int(2**(log(N_tot)//log(2)+1))
        timestep = float(time[1] - time[0])  # Sample step
        fsampt = 1.0 / timestep  # Sample frequency
        freqscale = N_tot / fsampt
        freqs = [i - int(N_tot / 2) for i in range(int(N_tot))]
        if is_real and is_time:
            # freqs = rfftfreq(N_tot, 1/(N_tot*fsampt))
            freqs = [i for i in range(int(N_tot / 2) + 1)]
            # freqs.append(-freqs[0])
        if is_time:
            freqs = [i / freqscale for i in freqs]
    return freqs


def comp_fft_time(freqs, is_angle, is_real):
    """Computes the time/space vector from the frequency/wavenumber vector
    Parameters
    ----------
    freqs: array
        Frequency or wavenumber vector
    is_angle: bool
        Boolean indicating if we output angle or time
    Returns
    -------
    Time/space vector
    """
    if freqs.size == 1:
        time = array([0])
    else:
        if is_real and not is_angle:
            N_tot = 2 * (len(freqs) - 1)  # Number of samples
            fs = freqs[-1] / (N_tot)
        else:
            N_tot = len(freqs)  # Number of samples
            fs = freqs[-1] / (N_tot - 2)
        tf = 1 / (fs * 2)
        time = linspace(0, tf, N_tot, endpoint=False)
        # fsampt = freqs[-1] * 2.0
        # timestep = 1.0 / fsampt
        if is_angle:
            time *= 2.0 * pi
            # timestep *= 2.0 * pi
        # time = [0 + i * timestep for i in range(N_tot)]
    return time.tolist()


def comp_nthoctave_axis(noct, freqmin, freqmax):
    """Computes the frequency vector between freqmin and freqmax for the 1/n octave
    Parameters
    ----------
    noct: int
        kind of octave band (1/3, etc)
    freqmin: float
        minimum frequency
    freqmax: float
        maximum frequency
    Returns
    -------
    Frequency vector
    """
    if noct == 3:
        table = [
            12.5,
            16,
            20,
            25,
            31.5,
            40,
            50,
            63,
            80,
            100,
            125,
            160,
            200,
            250,
            315,
            400,
            500,
            630,
            800,
            1000,
            1250,
            1600,
            2000,
            2500,
            3150,
            4000,
            5000,
            6300,
            8000,
            10000,
            12500,
            16000,
            20000,
        ]
        f_oct = [f for f in table if (f >= freqmin and f <= freqmax)]
    elif noct == 1:
        table = [
            16,
            31.5,
            63,
            125,
            250,
            500,
            1000,
            2000,
            4000,
            8000,
            16000,
        ]
        f_oct = [f for f in table if (f >= freqmin and f <= freqmax)]
    else:
        f0 = 1000
        f_oct = [f0]
        i = 1
        while f_oct[-1] <= freqmax:
            f = f0 * 2.0 ** (i / noct)
            f_oct.append(0.5 * ceil(2.0 * f))
            i = i + 1
        f_oct = f_oct[:-1]
        i = -1
        while f_oct[0] > freqmin:
            f = f0 * 2.0 ** (i / noct)
            f_oct = [0.5 * ceil(2.0 * f)] + f_oct
            i = i - 1
        f_oct = f_oct[1:]
    return f_oct


# def _comp_fft(values, is_positive=False):
#     """Computes the Fourier Transform
#     Parameters
#     ----------
#     values: ndarray
#         ndarray of the field
#     Returns
#     -------
#     Complex Fourier Transform
#     """
#     values_FT = fft(values)
#     if is_positive:
#         if iscomplex(values).any():
#             print("WARNING: keeping only positive harmonics from complex raw data")
#         values_FT[0] *= 0.5
#         values_FT = 2.0 * fftshift(values_FT) / len(values)
#         if len(values) % 2 == 0:
#             values_FT = append(values_FT, conjugate(values_FT[0]))
#     else:
#         values_FT = fftshift(values_FT) / len(values)
#     return values_FT


def comp_fftn(values, axes_list, is_real=True):
    """Computes the Fourier Transform
    Parameters
    ----------
    values: ndarray
        ndarray of the field
    Returns
    -------
    Complex Fourier Transform
    """

    axes = []
    axes_dict_non_uniform = dict()
    shape = []
    is_onereal = False
    is_twice = False
    axis_names = [axis.name for axis in axes_list]

    # Check if one or several axes is non-uniform
    is_non_uniform = False
    for axis in axes_list:
        if axis.transform == "fft":
            if axis.corr_values is not None:  # Timesteps
                if not is_uniform(axis.corr_values):
                    is_non_uniform = True
                    axes_dict_non_uniform[axis.index] = [
                        axis.corr_values,
                        axis.input_data,
                    ]
                    # Keep only interpolation data
                    axis.values = axis.input_data
                    axis.input_data = None
                    continue
            if axis.input_data is not None:
                if not isin(axis.input_data, axis.values).all():
                    is_non_uniform = True
                    # Convert wavenumbers to frequencies if needed
                    frequencies = (
                        axis.input_data
                        if axis.name == "freqs"
                        else axis.input_data / (2 * pi)
                    )
                    axes_dict_non_uniform[axis.index] = [
                        axis.corr_values,
                        frequencies,
                    ]
                    # Keep only interpolation data
                    axis.values = axis.input_data
                    axis.input_data = None

    # Apply DFT on non-uniform axes
    if is_non_uniform:
        values = nudftn(values, axes_dict=axes_dict_non_uniform)

    # Find other fft axes
    for axis in axes_list:
        if axis.index not in axes_dict_non_uniform:
            if axis.transform == "fft":
                if is_real and axis.name == "freqs":
                    axes.append(axis.index)
                    shape.append(values.shape[axis.index])
                    is_onereal = True
                elif (
                    is_real
                    and axis.name == "wavenumber"
                    and "freqs" not in axis_names
                    and min(axis.values) >= 0
                ):
                    axes.append(axis.index)
                    shape.append(values.shape[axis.index])
                    is_twice = True
                else:
                    axes = [axis.index] + axes
                    shape = [values.shape[axis.index]] + shape
    if axes != []:
        size = array(shape).prod()
        if is_onereal:
            values_FT = rfftn(values, axes=axes)
            # Do not multiply constant component by 2 (f=0)
            if axes_list[axes[-1]].corr_values is not None:
                freqs = axes_list[axes[-1]].corr_values
            else:
                freqs = axes_list[axes[-1]].values
            if freqs[0] == 0:
                slice_0 = take(values_FT, 0, axis=axes[-1])
                slice_0 *= 0.5
                if is_twice:
                    slice_0 *= 0.5
                other_values = delete(values_FT, 0, axis=axes[-1])
                values_FT = insert(other_values, 0, slice_0, axis=axes[-1])
            values_FT2 = 2.0 * fftshift(values_FT, axes=axes[:-1]) / size
            if is_twice:
                values_FT2 *= 2.0
        elif is_twice:
            values_FT = fftn(values, axes=axes)
            slice_0 = take(values_FT, 0, axis=axes[-1])
            slice_0 *= 0.5
            other_values = delete(values_FT, 0, axis=axes[-1])
            values_FT = insert(other_values, 0, slice_0, axis=axes[-1])
            values_FT2 = 2.0 * fftshift(values_FT, axes=axes) / size
        else:
            values_FT = fftn(values, axes=axes)
            values_FT2 = fftshift(values_FT, axes=axes) / size
    else:
        values_FT2 = values
    return values_FT2


# def _comp_ifft(values, is_positive=False):
#     """Computes the Inverse Fourier Transform
#     Parameters
#     ----------
#     values: ndarray
#         ndarray of the FT
#     Returns
#     -------
#     IFT
#     """

#     if is_positive:
#         values[0] *= 2
#         values = concatenate((flip(conjugate(values))[:-1], values))[:-1]
#         values = values / 2
#         values = ifftshift(values) * len(values)
#     else:
#         values = ifftshift(values) * len(values)
#     values_IFT = ifft(values)
#     return values_IFT


def comp_ifftn(values, axes_list, is_real=True):
    """Computes the Inverse Fourier Transform
    Parameters
    ----------
    values: ndarray
        ndarray of the FT
    Returns
    -------
    IFT
    """

    axes = []
    shape = []
    is_onereal = False
    is_half = False
    axis_names = [axis.name for axis in axes_list]
    # Check if one or several axes is non-uniform
    axes_dict_non_uniform = {}

    for axis in axes_list:
        if axis.transform == "ifft":
            if axis.input_data is not None:
                if not is_uniform(axis.input_data):
                    # Data is at least non uniform in "space"
                    axes_dict_non_uniform[axis.index] = [
                        axis.input_data,
                        axis.corr_values,
                    ]
                    # Keep only interpolation data
                    axis.values = axis.input_data
                    axis.input_data = None
                elif axis.corr_values is not None:
                    # Compare frequencies of interest with fft frequencies
                    freqs = comp_fft_freqs(
                        axis.input_data, axis.name == "time", is_real
                    )
                    if len(freqs) != len(axis.corr_values) or not allclose(
                        freqs,
                        axis.corr_values,
                        rtol=1e-5,
                        atol=1e-8,
                        equal_nan=False,
                    ):
                        # Data is at least non uniform in "frequency"
                        # Convert wavenumbers to frequencies if needed
                        frequencies = (
                            axis.corr_values
                            if axis.name == "time"
                            else axis.corr_values / (2 * pi)
                        )
                        axes_dict_non_uniform[axis.index] = [
                            axis.input_data,
                            frequencies,
                        ]
                        # Keep only interpolation data
                        axis.values = axis.input_data
                        axis.input_data = None
    # Compute non uniform inverse Fourier Transform for axes
    if axes_dict_non_uniform:
        values = inudftn(values, axes_dict=axes_dict_non_uniform)

    for axis in axes_list:
        # Exclude non uniform axes
        if axis.index in axes_dict_non_uniform:
            continue
        if axis.transform == "ifft":
            if is_real and axis.name == "time":
                axes.append(axis.index)
                shape.append(2 * (values.shape[axis.index] - 1))
                is_onereal = True
            elif is_real and axis.name == "angle" and "time" not in axis_names:
                axes.append(axis.index)
                shape.append(values.shape[axis.index])
                is_half = True
            else:
                axes = [axis.index] + axes
                shape = [values.shape[axis.index]] + shape
    if axes:  # Check if ifftn has to be called
        size = array(shape).prod()
        if is_onereal:
            values = values * size / 2
            if is_half:
                values *= 0.5
            values_shift = ifftshift(values, axes=axes[:-1])
            slice_0 = take(values_shift, 0, axis=axes[-1])
            slice_0 *= 2
            if is_half:
                slice_0 *= 2
            other_values = delete(values_shift, 0, axis=axes[-1])
            values = insert(other_values, 0, slice_0, axis=axes[-1])
            values_IFT = irfftn(values, axes=axes)
        elif is_half:
            values = values * size / 2
            values_shift = ifftshift(values, axes=axes[:-1])
            slice_0 = take(values_shift, 0, axis=axes[-1])
            slice_0 *= 2
            other_values = delete(values_shift, 0, axis=axes[-1])
            values = insert(other_values, 0, slice_0, axis=axes[-1])
            values_IFT = ifftn(values, axes=axes)
        else:
            values_shift = ifftshift(values, axes=axes) * size
            values_IFT = ifftn(values_shift, axes=axes)
    else:
        values_IFT = values
    if is_real:
        values_IFT = real(values_IFT)
    return values_IFT


def comp_magnitude(values):
    """Computes the magnitude of the Fourier Transform
    Parameters
    ----------
    values: ndarray
        ndarray of the field
    Returns
    -------
    Magnitude of the Fourier Transform
    """
    return np_abs(_comp_fft(values))


#    return np_abs(comp_stft_average(values))
def comp_phase(values):
    """Computes the phase of the Fourier Transform
    Parameters
    ----------
    values: ndarray
        ndarray of the field
    Returns
    -------
    Phase of the Fourier Transform
    """
    return np_angle(_comp_fft(values))


def comp_stft_average(values):
    """Computes the average of the Short-Time Fourier Transform
    Parameters
    ----------
    values: ndarray
        ndarray of the field
    Returns
    -------
    Average of the Short-Time Fourier Transform
    """
    # To do
    nperseg = 3200
    noverlap = int(nperseg * 0.75)
    f, t, Zxx = stft(
        values, fs=48000, window="hann", nperseg=nperseg, noverlap=noverlap
    )
    window_size = nperseg / len(values)
    values = mean(Zxx, axis=1) / (0.5)
    #    values = 2.0 * mean(Zxx, axis=1)
    print(values.shape)
    return f, np_abs(values)


def comp_fft_average(values):
    """Computes the average of the Short-Time Fourier Transform
    Parameters
    ----------
    values: ndarray
        ndarray of the field
    Returns
    -------
    Average of the Short-Time Fourier Transform
    """
    # To do
    nperseg = 3200
    noverlap = int(nperseg * 0.75)
    step = nperseg - noverlap
    N_tot = len(values)
    nwindows = int(N_tot / (2.0 * step))
    values_fft = zeros(nperseg, dtype="complex128")
    for i in range(nwindows):
        values_fft += _comp_fft(
            values[i * step : nperseg + i * step] * hanning(nperseg)
        )
    values = values_fft[int(nperseg / 2) :] / nwindows
    f = linspace(0, int(N_tot / 2), int(nperseg / 2))
    return f, np_abs(values)


def rect_window(f, M, dt):
    W = where(
        isclose(f, 0),
        (1 - exp(-1j * 2 * pi * dt * f * (M))) / (1 - exp(-1j * 2 * pi * dt * f)) / M,
        1,
    )
    return W


# M = 200
# tf = 1
# timec = linspace(0,tf*(1-1/M),M)
# dt = timec[1] - timec[0]
# A0 = 2
# freq0 = 10.0
# phi0=0
# y = A0*cos(2*pi*freq0*timec+phi0)
# freqs = comp_fft_freqs(timec, is_time=True, is_positive=False)
# y_FT = comp_fft(y)
# fig = plt.figure(constrained_layout=True, figsize=(20, 10))
# plt.plot(freqs,np_abs(y_FT))
# freqs_th = [-freq0, freq0]
# If = [0 for i in range(len(freqs_th))]
# for i in range(len(freqs_th)):
#     If[i] = int(argmin(abs([f-freqs_th[i] for f in freqs])))
# freqs = [freqs[i] for i in If]
# (xfreqs2, xfreqs1) = meshgrid(freqs_th,freqs)
# Wmat = rect_window(xfreqs1 - xfreqs2, M, dt)
# y_corr = linalg.solve(Wmat, y_FT[If])
# print(y_corr)
# plt.plot(freqs_th,np_abs(y_corr))



# -*- coding: utf-8 -*-

from os.path import isdir, join
from os import walk, getcwd, chdir
from .Load.load_json import load_json
from .Load.load_hdf5 import load_hdf5
from .Load.load_pkl import load_pkl
from .Load.import_class import import_class
from ..Classes._check import InitUnKnowClassError


def init_data(obj, file_path):
    """
    Initialize pyleecan objects (by init_dict) within list and/or dict data structure.
    Non pyleecan, list or dict type data will be kept as they are.

    Parameters
    ----------
    obj: object
        list/dict containing pyleecan init_dict data

    file_path: str
        path of the obj loaded

    Returns
    -------
    data:
        initialized pyleecan objects within a list or dict
    """

    # Find the folder path
    idx = max(file_path.rfind("/"), file_path.rfind("\\"))
    if idx == -1:
        folder_path = ""
    else:
        folder_path = file_path[: idx + 1]

    # --- list type ---
    if isinstance(obj, list):
        data = []
        for elem in obj:
            data.append(init_data(elem, file_path))
        return data
    # --- dict type ---
    if isinstance(obj, dict):
        # --- pyleecan class (has to be checked befor 'normal' dict) ---
        # Check if the dictionay has a "__class__" key
        if "__class__" in obj:
            # Check if data is a pyleecan class
            class_obj = import_class("pyleecan.Classes", obj.get("__class__"), "")
            if folder_path != "":
                wd = getcwd()
                chdir(folder_path)
                new_obj = class_obj(init_dict=obj)
                chdir(wd)
                return new_obj
            else:
                return class_obj(init_dict=obj)

        # --- 'normal' dict ---
        data = dict()
        for key in obj:
            data[key] = init_data(obj[key], file_path)
        return data

    # --- other type ---
    # keep all other (json) types as they are
    else:
        return obj


def load_init_dict(file_path):
    """load the init_dict from a h5 or json file"""
    if file_path.endswith("hdf5") or file_path.endswith("h5"):
        return load_hdf5(file_path)
    elif file_path.endswith("json") or isdir(file_path):
        return load_json(file_path)
    else:
        raise Exception(
            "Load error: Only hdf5, h5 and json format supported: " + file_path
        )


def load(file_path):
    """Load a pyleecan object from a json file

    Parameters
    ----------
    file_path: str
        path to the file to load
    """
    if file_path.endswith(".pkl"):
        return load_pkl(file_path)
    file_path, init_dict = load_init_dict(file_path)

    # Check that loaded data are of type dict
    if not isinstance(init_dict, dict):
        raise LoadWrongTypeError(
            'Loaded file is of type "'
            + type(init_dict).__name__
            + '", type "dict" expected.'
        )
    # Check that the dictionay has a "__class__" key
    if "__class__" not in init_dict:
        raise LoadWrongDictClassError('Key "__class__" missing in loaded file')

    return init_data(init_dict, file_path)


def _load(file_path, cls_type=None):
    """Load a list of pyleecan objects from a json file

    Parameters
    ----------
    file_path: str
        path to the file to load
    """
    _, obj = load_json(file_path)

    # check the initial object's type if set
    if cls_type is not None:
        if type(obj).__name__ != cls_type:
            raise LoadWrongTypeError(
                'Object is of type "'
                + type(obj).__name__
                + '", type "'
                + cls_type
                + '" expected.'
            )

    return init_data(obj, file_path)


def load_list(file_path):
    return _load(file_path, "list")


def load_dict(file_path):
    return _load(file_path, "dict")


def load_matlib(mat_path):
    """Load all the Material json file from a folder and subfolder

    Parameters
    ----------
    mat_path: str
        path to the file to load

    Returns
    -------
    matlib: list
        List of Material object
    """

    # Check that the dir exist
    if not isdir(mat_path):
        raise LoadMissingFolderError(
            "The following given path doesn't lead to a directory: " + mat_path
        )

    # Get and Read all the file to create a list dictionary : variable name <=> value
    matlib = list()
    for (dirpath, _, filenames) in walk(mat_path):
        for file_name in filenames:
            # For all json file in the folder and subfolder
            if file_name.endswith(".json") or file_name.endswith(".h5"):
                file_path = join(dirpath, file_name)
                try:
                    mat = load(file_path)
                    # Update the object property
                    mat.name = file_name[:-5]
                    mat.path = file_path
                    # Keep only the materials
                    if isinstance(mat, import_class("pyleecan.Classes", "Material")):
                        matlib.append(mat)
                except Exception:
                    print("When loading matlib, unable to load file: " + file_path)
    return matlib


class LoadMissingFolderError(Exception):
    """ """

    pass


class LoadWrongDictClassError(Exception):
    """ """

    pass


class LoadWrongTypeError(Exception):
    """ """

    pass


class LoadSwitchError(Exception):
    """ """

    pass



# -*- coding: utf-8 -*-
# File generated according to Generator/ClassesRef/RequestedAxis.csv
# WARNING! All changes made in this file will be lost!
"""Method code available at https://github.com/Eomys/SciDataTool/tree/master/SciDataTool/Methods//RequestedAxis
"""

from os import linesep
from sys import getsizeof
from ._check import set_array, check_var, raise_
from ..Functions.save import save
from ..Functions.copy import copy
from ..Functions.load import load_init_dict
from ..Functions.Load.import_class import import_class
from ._frozen import FrozenClass

# Import all class method
# Try/catch to remove unnecessary dependencies in unused method
try:
    from ..Methods.RequestedAxis.get_axis import get_axis
except ImportError as error:
    get_axis = error


from numpy import array, array_equal
from ._check import InitUnKnowClassError


class RequestedAxis(FrozenClass):
    """Class to handle requested axes during get_along methods"""

    VERSION = 1

    # cf Methods.RequestedAxis.get_axis
    if isinstance(get_axis, ImportError):
        get_axis = property(
            fget=lambda x: raise_(
                ImportError("Can't use RequestedAxis method get_axis: " + str(get_axis))
            )
        )
    else:
        get_axis = get_axis
    # save and copy methods are available in all object
    save = save
    copy = copy

    def __init__(
        self,
        name="",
        corr_name="",
        unit="",
        corr_unit="",
        extension="",
        values=None,
        indices=None,
        input_data=None,
        operation=None,
        index=None,
        transform=None,
        is_pattern=False,
        rebuild_indices=None,
        is_step=False,
        noct=None,
        corr_values=None,
        init_dict=None,
        init_str=None,
    ):
        """Constructor of the class. Can be use in three ways :
        - __init__ (arg1 = 1, arg3 = 5) every parameters have name and default values
            for SciDataTool type, -1 will call the default constructor
        - __init__ (init_dict = d) d must be a dictionnary with property names as keys
        - __init__ (init_str = s) s must be a string
        s is the file path to load

        ndarray or list can be given for Vector and Matrix
        object or dict can be given for SciDataTool Object"""

        if init_str is not None:  # Load from a file
            init_dict = load_init_dict(init_str)[1]
        if init_dict is not None:  # Initialisation by dict
            assert type(init_dict) is dict
            # Overwrite default value with init_dict content
            if "name" in list(init_dict.keys()):
                name = init_dict["name"]
            if "corr_name" in list(init_dict.keys()):
                corr_name = init_dict["corr_name"]
            if "unit" in list(init_dict.keys()):
                unit = init_dict["unit"]
            if "corr_unit" in list(init_dict.keys()):
                corr_unit = init_dict["corr_unit"]
            if "extension" in list(init_dict.keys()):
                extension = init_dict["extension"]
            if "values" in list(init_dict.keys()):
                values = init_dict["values"]
            if "indices" in list(init_dict.keys()):
                indices = init_dict["indices"]
            if "input_data" in list(init_dict.keys()):
                input_data = init_dict["input_data"]
            if "operation" in list(init_dict.keys()):
                operation = init_dict["operation"]
            if "index" in list(init_dict.keys()):
                index = init_dict["index"]
            if "transform" in list(init_dict.keys()):
                transform = init_dict["transform"]
            if "is_pattern" in list(init_dict.keys()):
                is_pattern = init_dict["is_pattern"]
            if "rebuild_indices" in list(init_dict.keys()):
                rebuild_indices = init_dict["rebuild_indices"]
            if "is_step" in list(init_dict.keys()):
                is_step = init_dict["is_step"]
            if "noct" in list(init_dict.keys()):
                noct = init_dict["noct"]
            if "corr_values" in list(init_dict.keys()):
                corr_values = init_dict["corr_values"]
        # Set the properties (value check and convertion are done in setter)
        self.parent = None
        self.name = name
        self.corr_name = corr_name
        self.unit = unit
        self.corr_unit = corr_unit
        self.extension = extension
        self.values = values
        self.indices = indices
        self.input_data = input_data
        self.operation = operation
        self.index = index
        self.transform = transform
        self.is_pattern = is_pattern
        self.rebuild_indices = rebuild_indices
        self.is_step = is_step
        self.noct = noct
        self.corr_values = corr_values

        # The class is frozen, for now it's impossible to add new properties
        self._freeze()

    def __str__(self):
        """Convert this object in a readeable string (for print)"""

        RequestedAxis_str = ""
        if self.parent is None:
            RequestedAxis_str += "parent = None " + linesep
        else:
            RequestedAxis_str += (
                "parent = " + str(type(self.parent)) + " object" + linesep
            )
        RequestedAxis_str += 'name = "' + str(self.name) + '"' + linesep
        RequestedAxis_str += 'corr_name = "' + str(self.corr_name) + '"' + linesep
        RequestedAxis_str += 'unit = "' + str(self.unit) + '"' + linesep
        RequestedAxis_str += 'corr_unit = "' + str(self.corr_unit) + '"' + linesep
        RequestedAxis_str += 'extension = "' + str(self.extension) + '"' + linesep
        RequestedAxis_str += (
            "values = "
            + linesep
            + str(self.values).replace(linesep, linesep + "\t")
            + linesep
            + linesep
        )
        RequestedAxis_str += (
            "indices = "
            + linesep
            + str(self.indices).replace(linesep, linesep + "\t")
            + linesep
        )
        RequestedAxis_str += (
            "input_data = "
            + linesep
            + str(self.input_data).replace(linesep, linesep + "\t")
            + linesep
            + linesep
        )
        RequestedAxis_str += 'operation = "' + str(self.operation) + '"' + linesep
        RequestedAxis_str += "index = " + str(self.index) + linesep
        RequestedAxis_str += 'transform = "' + str(self.transform) + '"' + linesep
        RequestedAxis_str += "is_pattern = " + str(self.is_pattern) + linesep
        RequestedAxis_str += (
            "rebuild_indices = "
            + linesep
            + str(self.rebuild_indices).replace(linesep, linesep + "\t")
            + linesep
            + linesep
        )
        RequestedAxis_str += "is_step = " + str(self.is_step) + linesep
        RequestedAxis_str += "noct = " + str(self.noct) + linesep
        RequestedAxis_str += (
            "corr_values = "
            + linesep
            + str(self.corr_values).replace(linesep, linesep + "\t")
            + linesep
            + linesep
        )
        return RequestedAxis_str

    def __eq__(self, other):
        """Compare two objects (skip parent)"""

        if type(other) != type(self):
            return False
        if other.name != self.name:
            return False
        if other.corr_name != self.corr_name:
            return False
        if other.unit != self.unit:
            return False
        if other.corr_unit != self.corr_unit:
            return False
        if other.extension != self.extension:
            return False
        if not array_equal(other.values, self.values):
            return False
        if other.indices != self.indices:
            return False
        if not array_equal(other.input_data, self.input_data):
            return False
        if other.operation != self.operation:
            return False
        if other.index != self.index:
            return False
        if other.transform != self.transform:
            return False
        if other.is_pattern != self.is_pattern:
            return False
        if not array_equal(other.rebuild_indices, self.rebuild_indices):
            return False
        if other.is_step != self.is_step:
            return False
        if other.noct != self.noct:
            return False
        if not array_equal(other.corr_values, self.corr_values):
            return False
        return True

    def compare(self, other, name="self"):
        """Compare two objects and return list of differences"""

        if type(other) != type(self):
            return ["type(" + name + ")"]
        diff_list = list()
        if other._name != self._name:
            diff_list.append(name + ".name")
        if other._corr_name != self._corr_name:
            diff_list.append(name + ".corr_name")
        if other._unit != self._unit:
            diff_list.append(name + ".unit")
        if other._corr_unit != self._corr_unit:
            diff_list.append(name + ".corr_unit")
        if other._extension != self._extension:
            diff_list.append(name + ".extension")
        if not array_equal(other.values, self.values):
            diff_list.append(name + ".values")
        if other._indices != self._indices:
            diff_list.append(name + ".indices")
        if not array_equal(other.input_data, self.input_data):
            diff_list.append(name + ".input_data")
        if other._operation != self._operation:
            diff_list.append(name + ".operation")
        if other._index != self._index:
            diff_list.append(name + ".index")
        if other._transform != self._transform:
            diff_list.append(name + ".transform")
        if other._is_pattern != self._is_pattern:
            diff_list.append(name + ".is_pattern")
        if not array_equal(other.rebuild_indices, self.rebuild_indices):
            diff_list.append(name + ".rebuild_indices")
        if other._is_step != self._is_step:
            diff_list.append(name + ".is_step")
        if other._noct != self._noct:
            diff_list.append(name + ".noct")
        if not array_equal(other.corr_values, self.corr_values):
            diff_list.append(name + ".corr_values")
        return diff_list

    def __sizeof__(self):
        """Return the size in memory of the object (including all subobject)"""

        S = 0  # Full size of the object
        S += getsizeof(self.name)
        S += getsizeof(self.corr_name)
        S += getsizeof(self.unit)
        S += getsizeof(self.corr_unit)
        S += getsizeof(self.extension)
        S += getsizeof(self.values)
        if self.indices is not None:
            for value in self.indices:
                S += getsizeof(value)
        S += getsizeof(self.input_data)
        S += getsizeof(self.operation)
        S += getsizeof(self.index)
        S += getsizeof(self.transform)
        S += getsizeof(self.is_pattern)
        S += getsizeof(self.rebuild_indices)
        S += getsizeof(self.is_step)
        S += getsizeof(self.noct)
        S += getsizeof(self.corr_values)
        return S

    def as_dict(self, **kwargs):
        """
        Convert this object in a json serializable dict (can be use in __init__).
        Optional keyword input parameter is for internal use only
        and may prevent json serializability.
        """

        RequestedAxis_dict = dict()
        RequestedAxis_dict["name"] = self.name
        RequestedAxis_dict["corr_name"] = self.corr_name
        RequestedAxis_dict["unit"] = self.unit
        RequestedAxis_dict["corr_unit"] = self.corr_unit
        RequestedAxis_dict["extension"] = self.extension
        if self.values is None:
            RequestedAxis_dict["values"] = None
        else:
            RequestedAxis_dict["values"] = self.values.tolist()
        RequestedAxis_dict["indices"] = (
            self.indices.copy() if self.indices is not None else None
        )
        if self.input_data is None:
            RequestedAxis_dict["input_data"] = None
        else:
            RequestedAxis_dict["input_data"] = self.input_data.tolist()
        RequestedAxis_dict["operation"] = self.operation
        RequestedAxis_dict["index"] = self.index
        RequestedAxis_dict["transform"] = self.transform
        RequestedAxis_dict["is_pattern"] = self.is_pattern
        if self.rebuild_indices is None:
            RequestedAxis_dict["rebuild_indices"] = None
        else:
            RequestedAxis_dict["rebuild_indices"] = self.rebuild_indices.tolist()
        RequestedAxis_dict["is_step"] = self.is_step
        RequestedAxis_dict["noct"] = self.noct
        if self.corr_values is None:
            RequestedAxis_dict["corr_values"] = None
        else:
            RequestedAxis_dict["corr_values"] = self.corr_values.tolist()
        # The class name is added to the dict for deserialisation purpose
        RequestedAxis_dict["__class__"] = "RequestedAxis"
        return RequestedAxis_dict

    def _set_None(self):
        """Set all the properties to None (except SciDataTool object)"""

        self.name = None
        self.corr_name = None
        self.unit = None
        self.corr_unit = None
        self.extension = None
        self.values = None
        self.indices = None
        self.input_data = None
        self.operation = None
        self.index = None
        self.transform = None
        self.is_pattern = None
        self.rebuild_indices = None
        self.is_step = None
        self.noct = None
        self.corr_values = None

    def _get_name(self):
        """getter of name"""
        return self._name

    def _set_name(self, value):
        """setter of name"""
        check_var("name", value, "str")
        self._name = value

    name = property(
        fget=_get_name,
        fset=_set_name,
        doc=u"""Name requested in get_along command

        :Type: str
        """,
    )

    def _get_corr_name(self):
        """getter of corr_name"""
        return self._corr_name

    def _set_corr_name(self, value):
        """setter of corr_name"""
        check_var("corr_name", value, "str")
        self._corr_name = value

    corr_name = property(
        fget=_get_corr_name,
        fset=_set_corr_name,
        doc=u"""Corresponding axis if a transform is required

        :Type: str
        """,
    )

    def _get_unit(self):
        """getter of unit"""
        return self._unit

    def _set_unit(self, value):
        """setter of unit"""
        check_var("unit", value, "str")
        self._unit = value

    unit = property(
        fget=_get_unit,
        fset=_set_unit,
        doc=u"""Unit requested in get_along command

        :Type: str
        """,
    )

    def _get_corr_unit(self):
        """getter of corr_unit"""
        return self._corr_unit

    def _set_corr_unit(self, value):
        """setter of corr_unit"""
        check_var("corr_unit", value, "str")
        self._corr_unit = value

    corr_unit = property(
        fget=_get_corr_unit,
        fset=_set_corr_unit,
        doc=u"""Corresponding unit if a transform is required

        :Type: str
        """,
    )

    def _get_extension(self):
        """getter of extension"""
        return self._extension

    def _set_extension(self, value):
        """setter of extension"""
        check_var("extension", value, "str")
        self._extension = value

    extension = property(
        fget=_get_extension,
        fset=_set_extension,
        doc=u"""Extension of the requested axis (single or interval)

        :Type: str
        """,
    )

    def _get_values(self):
        """getter of values"""
        return self._values

    def _set_values(self, value):
        """setter of values"""
        if type(value) is int and value == -1:
            value = array([])
        elif type(value) is list:
            try:
                value = array(value)
            except:
                pass
        check_var("values", value, "ndarray")
        self._values = value

    values = property(
        fget=_get_values,
        fset=_set_values,
        doc=u"""Values of the axis

        :Type: ndarray
        """,
    )

    def _get_indices(self):
        """getter of indices"""
        return self._indices

    def _set_indices(self, value):
        """setter of indices"""
        if type(value) is int and value == -1:
            value = list()
        check_var("indices", value, "list")
        self._indices = value

    indices = property(
        fget=_get_indices,
        fset=_set_indices,
        doc=u"""Indices of the axis

        :Type: list
        """,
    )

    def _get_input_data(self):
        """getter of input_data"""
        return self._input_data

    def _set_input_data(self, value):
        """setter of input_data"""
        if type(value) is int and value == -1:
            value = array([])
        elif type(value) is list:
            try:
                value = array(value)
            except:
                pass
        check_var("input_data", value, "ndarray")
        self._input_data = value

    input_data = property(
        fget=_get_input_data,
        fset=_set_input_data,
        doc=u"""Input data to interpolate on

        :Type: ndarray
        """,
    )

    def _get_operation(self):
        """getter of operation"""
        return self._operation

    def _set_operation(self, value):
        """setter of operation"""
        check_var("operation", value, "str")
        self._operation = value

    operation = property(
        fget=_get_operation,
        fset=_set_operation,
        doc=u"""Operation to perform on the axis (coordinate change, etc)

        :Type: str
        """,
    )

    def _get_index(self):
        """getter of index"""
        return self._index

    def _set_index(self, value):
        """setter of index"""
        check_var("index", value, "int")
        self._index = value

    index = property(
        fget=_get_index,
        fset=_set_index,
        doc=u"""Index of the axis in the axes list

        :Type: int
        """,
    )

    def _get_transform(self):
        """getter of transform"""
        return self._transform

    def _set_transform(self, value):
        """setter of transform"""
        check_var("transform", value, "str")
        self._transform = value

    transform = property(
        fget=_get_transform,
        fset=_set_transform,
        doc=u"""Transform to perform on the axis (fft, ifft)

        :Type: str
        """,
    )

    def _get_is_pattern(self):
        """getter of is_pattern"""
        return self._is_pattern

    def _set_is_pattern(self, value):
        """setter of is_pattern"""
        check_var("is_pattern", value, "bool")
        self._is_pattern = value

    is_pattern = property(
        fget=_get_is_pattern,
        fset=_set_is_pattern,
        doc=u"""To indicate if the axis is a DataPattern

        :Type: bool
        """,
    )

    def _get_rebuild_indices(self):
        """getter of rebuild_indices"""
        return self._rebuild_indices

    def _set_rebuild_indices(self, value):
        """setter of rebuild_indices"""
        if type(value) is int and value == -1:
            value = array([])
        elif type(value) is list:
            try:
                value = array(value)
            except:
                pass
        check_var("rebuild_indices", value, "ndarray")
        self._rebuild_indices = value

    rebuild_indices = property(
        fget=_get_rebuild_indices,
        fset=_set_rebuild_indices,
        doc=u"""Indices to rebuild pattern

        :Type: ndarray
        """,
    )

    def _get_is_step(self):
        """getter of is_step"""
        return self._is_step

    def _set_is_step(self, value):
        """setter of is_step"""
        check_var("is_step", value, "bool")
        self._is_step = value

    is_step = property(
        fget=_get_is_step,
        fset=_set_is_step,
        doc=u"""To indicate if the pattern axis is step (for interpolation)

        :Type: bool
        """,
    )

    def _get_noct(self):
        """getter of noct"""
        return self._noct

    def _set_noct(self, value):
        """setter of noct"""
        check_var("noct", value, "int")
        self._noct = value

    noct = property(
        fget=_get_noct,
        fset=_set_noct,
        doc=u"""To store 1/nth octave band

        :Type: int
        """,
    )

    def _get_corr_values(self):
        """getter of corr_values"""
        return self._corr_values

    def _set_corr_values(self, value):
        """setter of corr_values"""
        if type(value) is int and value == -1:
            value = array([])
        elif type(value) is list:
            try:
                value = array(value)
            except:
                pass
        check_var("corr_values", value, "ndarray")
        self._corr_values = value

    corr_values = property(
        fget=_get_corr_values,
        fset=_set_corr_values,
        doc=u"""To store original axis values (useful in case of non uniform fft)

        :Type: ndarray
        """,
    )



# -*- coding: utf-8 -*-

from numpy import zeros

from SciDataTool.Functions.conversions import xy_to_rphi, cart2pol
from SciDataTool.Functions import AxisError


def get_rphiz_along(
    self, *args, unit="SI", is_norm=False, axis_data=[], is_squeeze=True
):
    """Returns the list of the cylindrical (r,phi,z) components of the field, using conversions and symmetries if needed.
    Parameters
    ----------
    self: Data
        a Data object
    *args: list of strings
        List of axes requested by the user, their units and values (optional)
    unit: str
        Unit requested by the user ("SI" by default)
    is_norm: bool
        Boolean indicating if the field must be normalized (False by default)
    axis_data: list
        list of ndarray corresponding to user-input data
    Returns
    -------
    list of 1Darray of axes values, ndarray of field values
    """

    if len(args) == 1 and type(args[0]) == tuple:
        args = args[0]  # if called from another script with *args

    if "comp_x" in self.components.keys() and "comp_y" in self.components.keys():
        # Extract from DataND
        resultx = self.components["comp_x"].get_along(
            args, unit=unit, is_norm=is_norm, axis_data=axis_data, is_squeeze=is_squeeze
        )
        resulty = self.components["comp_y"].get_along(
            args, unit=unit, is_norm=is_norm, axis_data=axis_data, is_squeeze=is_squeeze
        )
        field_x = resultx[self.components["comp_x"].symbol]
        field_y = resulty[self.components["comp_y"].symbol]
        shape = field_x.shape
        if "angle" not in resultx:
            raise AxisError(
                "ERROR: need angle axis to convert to cylindrical coordinates"
            )
        phi = resultx["angle"]
        # Convert to cylindrical coordinates
        (field_r, field_c) = cart2pol(field_x, field_y, phi)
        if "axial" in self.components.keys():
            resultz = self.components["axial"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_z = resultz[self.components["axial"].symbol]
        elif "comp_z" in self.components.keys():
            resultz = self.components["comp_z"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_z = resultz[self.components["comp_z"].symbol]
        else:
            field_z = zeros(shape)
        return_dict = dict(resultx)
        del return_dict[self.components["comp_x"].symbol]

    elif "radial" in self.components.keys():
        resultr = self.components["radial"].get_along(
            args, unit=unit, is_norm=is_norm, axis_data=axis_data, is_squeeze=is_squeeze
        )
        field_r = resultr[self.components["radial"].symbol]
        shape = field_r.shape
        if "tangential" in self.components.keys():
            resultphi = self.components["tangential"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_c = resultphi[self.components["tangential"].symbol]
        else:
            field_c = zeros(shape)
        if "axial" in self.components.keys():
            resultz = self.components["axial"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_z = resultz[self.components["axial"].symbol]
        elif "comp_z" in self.components.keys():
            resultz = self.components["comp_z"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_z = resultz[self.components["comp_z"].symbol]
        else:
            field_z = zeros(shape)
        return_dict = dict(resultr)
        del return_dict[self.components["radial"].symbol]

    elif "tangential" in self.components.keys():
        resultphi = self.components["tangential"].get_along(
            args, unit=unit, is_norm=is_norm, axis_data=axis_data, is_squeeze=is_squeeze
        )
        field_c = resultphi[self.components["tangential"].symbol]
        shape = field_c.shape
        if "radial" in self.components.keys():
            resultr = self.components["radial"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_r = resultphi[self.components["radial"].symbol]
        else:
            field_r = zeros(shape)
        if "axial" in self.components.keys():
            resultz = self.components["axial"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_z = resultz[self.components["axial"].symbol]
        elif "comp_z" in self.components.keys():
            resultz = self.components["comp_z"].get_along(
                args,
                unit=unit,
                is_norm=is_norm,
                axis_data=axis_data,
                is_squeeze=is_squeeze,
            )
            field_z = resultz[self.components["comp_z"].symbol]
        else:
            field_z = zeros(shape)
        return_dict = dict(resultphi)
        del return_dict[self.components["tangential"].symbol]

    elif "axial" in self.components.keys():
        resultz = self.components["axial"].get_along(
            args, unit=unit, is_norm=is_norm, axis_data=axis_data, is_squeeze=is_squeeze
        )
        field_z = resultz[self.components["axial"].symbol]
        shape = field_z.shape
        field_r = zeros(shape)
        field_c = zeros(shape)
        return_dict = dict(resultz)
        del return_dict[self.components["axial"].symbol]

    elif "comp_z" in self.components.keys():
        resultz = self.components["comp_z"].get_along(
            args, unit=unit, is_norm=is_norm, axis_data=axis_data, is_squeeze=is_squeeze
        )
        field_z = resultz[self.components["comp_z"].symbol]
        shape = field_z.shape
        field_r = zeros(shape)
        field_c = zeros(shape)
        return_dict = resultz
        del return_dict[self.components["comp_z"].symbol]

    else:
        raise AxisError(
            "Vector_field object is empty (should contain at least radial, tangential, axial, comp_x, comp_y or comp_z"
        )

    return_dict["radial"] = field_r
    return_dict["tangential"] = field_c
    return_dict["axial"] = field_z

    return return_dict



from numpy import column_stack, array, max as np_max
from SciDataTool.Functions.Plot.plot_2D import plot_2D
from SciDataTool.Functions.conversions import rphiz_to_xyz


def plot_2D_Data(
    self,
    *arg_list,
    axis_data=None,
    radius=None,
    is_norm=False,
    unit="SI",
    component_list=None,
    data_list=[],
    legend_list=[],
    color_list=None,
    curve_colors=None,
    phase_colors=None,
    linestyles=None,
    linewidth_list=[2],
    save_path=None,
    x_min=None,
    x_max=None,
    y_min=None,
    y_max=None,
    is_logscale_x=False,
    is_logscale_y=False,
    is_disp_title=True,
    is_grid=True,
    is_auto_ticks=True,
    is_auto_range=False,
    fig=None,
    ax=None,
    barwidth=100,
    type_plot=None,
    fund_harm_dict=None,
    is_show_fig=None,
    win_title=None,
    thresh=None,
    font_name="arial",
    font_size_title=12,
    font_size_label=10,
    font_size_legend=8,
    scale_units="x",
    scale=None,
    width=0.005,
):
    """Plots a field as a function of time

    Parameters
    ----------
    self : Output
        an Output object
    Data_str : str
        name of the Data Object to plot (e.g. "mag.Br")
    *arg_list : list of str
        arguments to specify which axes to plot
    radius : float
        radius used for quiver plot
    is_norm : bool
        boolean indicating if the field must be normalized
    unit : str
        unit in which to plot the field
    data_list : list
        list of Data objects to compare
    component_list : list
        list of component names to plot in separate figures
    legend_list : list
        list of legends to use for each Data object (including reference one) instead of data.name
    color_list : list
        list of colors to use for each Data object
    save_path : str
        full path of the png file where the figure is saved if save_path is not None
    x_min : float
        minimum value for the x-axis
    x_max : float
        maximum value for the x-axis
    y_min : float
        minimum value for the y-axis
    y_max : float
        maximum value for the y-axis
    is_logscale_x : bool
        boolean indicating if the x-axis must be set in logarithmic scale
    is_logscale_y : bool
        boolean indicating if the y-axis must be set in logarithmic scale
    is_disp_title : bool
        boolean indicating if the title must be displayed
    is_grid : bool
        boolean indicating if the grid must be displayed
    is_auto_ticks : bool
        in fft, adjust ticks to freqs (deactivate if too close)
    is_auto_range : bool
        in fft, display up to 1% of max
    fig : Matplotlib.figure.Figure
        existing figure to use if None create a new one
    ax : Matplotlib.axes.Axes object
        ax on which to plot the data
    barwidth : float
        barwidth scaling factor, only if type_plot = "bargraph"
    type_plot : str
        type of 2D graph : "curve", "bargraph", "barchart" or "quiver"
    fund_harm_dict : dict
        Dict containing axis name as key and frequency/order/wavenumber of fundamental harmonic as value to display fundamental harmonic in red in the fft
    is_show_fig : bool
        True to show figure after plot
    thresh : float
        threshold for automatic fft ticks
    scale_units : str
        arrow lenght scale factor reference {'width', 'height', 'dots', 'pouces', 'x', 'y', 'xy'}
    scale : float
        arrow length factor
    width : float
        arrow width factor
    """

    # Special case of quiver plot
    if type_plot == "quiver":
        result = self.get_xyz_along(
            arg_list, axis_data=axis_data, unit=unit, is_norm=is_norm
        )
        if "x" in result and "y" in result:
            Xdatas = column_stack((result["x"], result["y"]))
        else:
            if radius is None:
                radius = 1
            phi = result["angle"]
            rphi = column_stack((array([radius] * len(phi)), phi))
            Xdatas = rphiz_to_xyz(rphi)
        Ydatas = column_stack((result["comp_x"], result["comp_y"]))
        # Normalize Ydatas
        # Ydatas = Ydatas / (np_max(Ydatas) * 0.001 * radius)
        plot_2D(
            [Xdatas],
            [Ydatas],
            color_list=color_list,
            linestyle_list=linestyles,
            linewidth_list=linewidth_list,
            title=self.name.capitalize() + " quiver plot",
            legend_list=[self.name],
            xlabel="[m]",
            ylabel="[m]",
            fig=fig,
            ax=ax,
            type_plot="quiver",
            save_path=save_path,
            is_show_fig=is_show_fig,
            win_title=win_title,
            font_name=font_name,
            font_size_title=font_size_title,
            font_size_label=font_size_label,
            font_size_legend=font_size_legend,
            is_grid=False,
            scale_units=scale_units,
            scale=scale,
            width=width,
        )

    else:

        # Call the plot on each component
        if component_list is None:  # default: extract all components
            component_list = self.components.keys()
        for i, comp in enumerate(component_list):
            # (fig, axes, patch_leg, label_leg) = init_fig(None, shape="rectangle")

            if save_path is not None and len(component_list) > 1:
                save_path_comp = (
                    save_path.split(".")[0] + "_" + comp + "." + save_path.split(".")[1]
                )
            else:
                save_path_comp = save_path

            self.components[comp].plot_2D_Data(
                arg_list,
                axis_data=axis_data,
                is_norm=is_norm,
                unit=unit,
                data_list=[dat.components[comp] for dat in data_list],
                legend_list=legend_list,
                color_list=color_list,
                save_path=save_path_comp,
                x_min=x_min,
                x_max=x_max,
                y_min=y_min,
                y_max=y_max,
                is_logscale_x=is_logscale_x,
                is_logscale_y=is_logscale_y,
                is_disp_title=is_disp_title,
                is_grid=is_grid,
                is_auto_ticks=is_auto_ticks,
                is_auto_range=is_auto_range,
                fig=fig,
                ax=ax,
                barwidth=barwidth,
                type_plot=type_plot,
                fund_harm_dict=fund_harm_dict,
                is_show_fig=is_show_fig,
                thresh=thresh,
                font_name=font_name,
                font_size_title=font_size_title,
                font_size_label=font_size_label,
                font_size_legend=font_size_legend,
            )



# -*- coding: utf-8 -*-
from SciDataTool.Functions.symmetries import (
    rebuild_symmetries as rebuild_symmetries_fct,
)

from numpy import take


def rebuild_symmetries(
    self,
    values,
    axes_list,
):
    """Reconstructs the field of a Data object taking symmetries into account
    Parameters
    ----------
    self: Data
        a Data object
    values: ndarray
        ndarray of a field
    axes_list: list
        a list of RequestedAxis objects
    Returns
    -------
    ndarray of the reconstructed field
    """

    for axis in axes_list:
        if (
            axis.transform != "fft"
            and axis.is_pattern
            and (
                axis.extension
                not in ["sum", "rss", "mean", "rms", "integrate", "smallestperiod"]
                and axis.indices is None
            )
        ):
            values = take(values, axis.rebuild_indices, axis.index)
        elif axis.transform != "fft" and axis.extension in [
            "whole",
            "interval",
            "oneperiod",
            "antiperiod",
            "smallestperiod",
        ]:
            if axis.extension == "smallestperiod":
                is_smallestperiod = True
                is_oneperiod = False
                is_antiperiod = False
            elif axis.extension == "antiperiod":
                is_smallestperiod = False
                is_oneperiod = False
                is_antiperiod = True
            elif axis.extension == "oneperiod":
                is_smallestperiod = False
                is_oneperiod = True
                is_antiperiod = False
            # Ignore symmetries if fft axis
            elif axis.name == "freqs" or axis.name == "wavenumber":
                is_smallestperiod = True
                is_oneperiod = False
                is_antiperiod = False
            else:
                is_smallestperiod = False
                is_oneperiod = False
                is_antiperiod = False

            # Rebuild symmetries
            axis_symmetries = self.axes[axis.index].symmetries
            if is_oneperiod:
                if "antiperiod" in axis_symmetries:
                    nper = axis_symmetries["antiperiod"]
                    axis_symmetries["antiperiod"] = 2
                    values = rebuild_symmetries_fct(values, axis.index, axis_symmetries)
                    axis_symmetries["antiperiod"] = nper
            elif not is_smallestperiod and not is_antiperiod:
                values = rebuild_symmetries_fct(values, axis.index, axis_symmetries)
    return values



from ...Classes._check import InitUnKnowClassError


def import_class(mod_name, class_name, prop_name=""):
    """Dynamical import of a class (for init_dict=class.as_dict())"""

    try:
        module = __import__(mod_name + "." + class_name, fromlist=[class_name])
        return getattr(module, class_name)
    except:
        if prop_name != "":
            raise InitUnKnowClassError(
                "Unknow class name "
                + class_name
                + " in init_dict for property "
                + prop_name
            )
        else:
            raise InitUnKnowClassError(
                "Unknow class name " + class_name + " when loading"
            )
